import time
import pygame
from cooptools.dictPolicies import TogglePolicy
from cooptools.sectors.grids import RectGrid
from cooptools.dictPolicies import IActOnDictPolicy
from coopgraph.gridgraph import GridGraph
from coopgame.grids_graphs.pygridhandler import PyGridHandler
from coopgame.grids_graphs.pygraphhandler import PyGraphHandler
from cooptools.toggles import BooleanToggleable
import numpy as np
from cooptools.colors import Color
from coopstructs.geometry.vectors.vectorN import Vector2
from typing import Callable, Any, Dict, Type, List
import logging
from coopgame.enums import GridDrawType
from coopstructs.geometry import Line, Rectangle
import uuid
from coopstructs.geometry.curves.curves import LineCurve
from coopgraph.graphs import AStarResults
import coopgame.pygamehelpers as help
from coopgame.linedrawing.pyCurveHandler import PyCurveHandler

class MouseGridResolverArgs:
    def __init__(self,
                 windowRect: Rectangle,
                 mousePos: Vector2,
                 color: Color):
        self.windowRect = windowRect
        self.mousePos = mousePos
        self.color = color


class PathFinderFactory():
    def __init__(self,
                 grid_graph: GridGraph,
                 astar_start_point: Vector2 = None,
                 wait_timer_ms: int = None):

        if wait_timer_ms is None:
            wait_timer_ms = 0

        super().__init__()
        self.astar_start_point = astar_start_point
        self.astar_result = None
        self.grid_graph = grid_graph
        self.wait_timer_ms = wait_timer_ms
        self.start_timer = None
        self.last_calc = None

    def calculate_astar_from_cached_pos(self,
                                        new_pos: Vector2):
        self.astar_result = self.grid_graph.astar_between_grid_pos(self.astar_start_point, new_pos)
        self.last_calc = time.perf_counter()
        logging.info(f"Route calculated from {self.astar_start_point}-->{new_pos}")

    def calculate_if_passed_threshold(self, new_pos: Vector2):
        if self.last_calc is not None:
            return

        if (self.wait_timer_ms is None) or \
                (self.start_timer and time.perf_counter() - self.start_timer > self.wait_timer_ms / 1000):
            self.calculate_astar_from_cached_pos(new_pos)

    def set_astar_origin(self, grid_coords: Vector2):
        self.astar_start_point = grid_coords
        self.astar_result = None

    def set_start_timer(self):
        self.last_calc = None
        self.astar_result = None
        self.start_timer = time.perf_counter()


def astar_curves(surface_rect: Rectangle,
                 grid_graph: GridGraph,
                 astar_results: AStarResults,
                 ):
    if astar_results and astar_results.path is not None:
        astr_curves = []
        last = None
        for ii in range(0, len(astar_results.path)):
            if last is not None:
                astr_curves.append(LineCurve(str(uuid.uuid4()),
                                             grid_graph.grid.coord_from_grid_pos(last.pos, surface_rect),
                                             grid_graph.grid.coord_from_grid_pos(astar_results.path[ii].focal_point,
                                                                                 surface_rect)))
            last = astar_results.path[ii]

        return astr_curves
    else:
        return None


def toggled_grids(grid: Type[RectGrid], toggle_key):
    grid_value_array = grid.state_value_as_array(keys=[toggle_key])
    toggleds = []
    for x in range(0, grid.nColumns):
        for y in range(0, grid.nRows):
            if grid_value_array[y][x].value:
                toggleds.append(Vector2(x, y))
    return toggleds


def define_highlighted_grids(grid: RectGrid,
                             predefined_highlights: Dict[Vector2, Color] = None,
                             highlight_toggled: bool = False,
                             toggle_key: Any = None):
    highlights = {}

    if highlight_toggled and toggle_key:
        for toggled in toggled_grids(grid, toggle_key):
            highlights[toggled] = Color.ORANGE

    if predefined_highlights:
        highlights.update(predefined_highlights)

    return highlights


class PyGridGraphHandler():

    @classmethod
    def draw_calculated_path(cls,
                             grid_graph: GridGraph,
                             astar_results: AStarResults,
                             path_color: Color,
                             window_rect: Rectangle,
                             surface: pygame.Surface,
                             rotation_matrix: np.ndarray,
                             path_node_color: Color = None):

        PyCurveHandler.draw_curves({x: path_color for x in astar_curves(window_rect,
                                                                        grid_graph,
                                                                        astar_results) or []},
                                   surface,
                                   draw_scale_matrix=rotation_matrix,
                                   control_point_color=path_node_color
                                   )

    def __init__(self,
                 x: int = None,
                 y: int = None,
                 init_condition: Dict[Any, Any] = None):
        super().__init__()
        self.toggle_key = 'toggle'
        self.toggle_bool_policy = TogglePolicy(key=self.toggle_key, toggle=BooleanToggleable(default=False))
        self.show_toggled = BooleanToggleable(default=False)
        self.allow_diag = BooleanToggleable(default=True,
                                            on_toggle_callbacks=[
                                                lambda x: self.grid_graph.toggle_allow_diagonal_connections('diag'),
                                                lambda x: self.pff.set_start_timer()])

        self.grid_graph: GridGraph = None
        self.pff = None

        if x and y:
            self.init_grid(x, y, init_condition)

        self.grid_handler = PyGridHandler(screen=self.screen,
                                          grid_provider=lambda: self.grid_graph.grid,
                                          get_game_area_rect_callback=self.get
                                          )
        self.graph_handler = PyGraphHandler(self.grid_graph.graph)

        self.graph_draw_invalidated: bool = True

        self._invalidated_highlights = True
        self._cached_highlighted_static = {}
        self._predefined_highlights = {}

    def highlighted_static(self, predefined_highlights):
        # if self._invalidated_highlights:
        #     self._cached_highlighted_static = define_highlighted_grids(self.grid, predefined_highlights, self.show_toggled.value, toggle_key=self.toggle_key)
        #     self._invalidated_highlights = False
        self._cached_highlighted_static = define_highlighted_grids(self.grid, predefined_highlights,
                                                                   self.show_toggled.value, toggle_key=self.toggle_key)
        return self._cached_highlighted_static

    @property
    def grid(self):
        if self.grid_graph is not None:
            return self.grid_graph.grid

    @property
    def graph(self):
        if self.grid_graph is not None:
            return self.grid_graph.graph

    def __getitem__(self, item):
        if item is None:
            return None

        if type(item) != Vector2:
            raise TypeError(f"Type item must be Vector2. {type(item)} was provided")
        return self.grid.at(item.y, item.x)[self.toggle_key]

    def init_grid(self,
                  rows,
                  cols,
                  init_condition: np.ndarray = None,
                  path_finder_delay_ms: int = None):
        grid = RectGrid(rows, cols)

        self.grid_graph = GridGraph(grid)

        if init_condition is not None:
            for ii, row in enumerate(init_condition):
                for jj, col in enumerate(row):
                    self.grid_graph.grid.at(ii, jj).state = init_condition

        self.pff = PathFinderFactory(self.grid_graph, wait_timer_ms=path_finder_delay_ms)

    def mouse_grid_pos(self, window_rect: Rectangle, mouse_pos: Vector2, draw_scale_matrix: np.array = None):
        return self.grid_handler.get_mouse_grid_pos(window_rect,
                                                    mouse_pos,
                                                    coord_to_grid_converter=self.grid_graph.grid.grid_from_coord,
                                                    draw_scale_matrix=draw_scale_matrix)

    def redefine_grid(self, rows: int, cols: int, grid_update_callback: Callable[[Any], Any] = None):
        if rows is None or cols is None:
            return

        logging.info(f"Creating grid...")
        self.init_grid(rows, cols)
        logging.info(f"Done creating grid...")

        if grid_update_callback is not None:
            grid_update_callback(...)

    def _resolve_predefined_highlights(self,
                                       highlight_mouse_pos: MouseGridResolverArgs = None,
                                       predefined_highlights: Dict[Vector2, Color] = None,
                                       rotation_matrix=None):

        if predefined_highlights == None and highlight_mouse_pos:
            pd_highlights = {}
        else:
            pd_highlights = predefined_highlights.copy()

        if highlight_mouse_pos:
            mouse_grid = self.mouse_grid_pos(highlight_mouse_pos.windowRect, highlight_mouse_pos.mousePos,
                                             draw_scale_matrix=rotation_matrix)

            if mouse_grid is not None: pd_highlights[mouse_grid] = highlight_mouse_pos.color

        return pd_highlights

    def _define_highlighted_grids(self, predefined_highlights):
        return self.highlighted_static(predefined_highlights, highlight_toggled=self.show_toggled.value)
        # return define_highlighted_grids(grid=self.grid_graph.grid,
        #                                   predefined_highlights=predefined_highlights,
        #                                   highlight_toggled=self.show_toggled.value,
        #                                   toggle_key=self.toggle_key)

    def draw(self,
             grid_surface: pygame.Surface = None,
             graph_surface: pygame.Surface = None,
             overlay_surface: pygame.Surface = None,
             # hover_pos_surface: pygame.Surface = None,
             grid_draw_type: GridDrawType = None,
             rotation_matrix: np.ndarray = None,
             predefined_highlights: Dict[Vector2, Color] = None,
             highlight_mouse_pos: MouseGridResolverArgs = None,
             grid_color: Color = None,
             ):

        if grid_draw_type is None:
            grid_draw_type = GridDrawType.LINES

        # draw grid base
        if grid_surface:
            if grid_color is None: grid_color = Color.DIM_GRAY
            self.grid_handler.draw_base_to_surface(grid_surface,
                                                   self.grid_graph.grid,
                                                   grid_color=grid_color,
                                                   grid_draw_type=grid_draw_type,
                                                   draw_scale_matrix=rotation_matrix)

        # draw graph
        if self.grid_graph.graph and graph_surface:
            coord_converter = lambda p: Vector2.from_tuple(self.grid_graph.grid.coord_from_grid_pos(p.as_tuple(), graph_surface.get_rect()))
            self.graph_handler.draw_to_surface(graph_surface,
                                               self.grid_graph.graph,
                                               draw_scale_matrix=rotation_matrix,
                                               coordinate_converter=coord_converter)

        # draw overlay
        if overlay_surface:
            predefined_highlights = self._resolve_predefined_highlights(highlight_mouse_pos,
                                                                        predefined_highlights,
                                                                        rotation_matrix=rotation_matrix)

            highlights = self.highlighted_static(predefined_highlights)

            self.grid_handler.draw_overlay_to_surface(overlay_surface,
                                                      self.grid_graph.grid,
                                                      additional_highlight_grid_cells=highlights,
                                                      draw_scale_matrix=rotation_matrix,

                                                      )
        # if hover_pos_surface:
        #     self.redraw_hover_grid_pos_overlay(hover_grid_surface=hover_pos_surface,
        #                                        highlight_mouse_grid_pos_color=hi)

    def handle_hover_over(self, window_rect: Rectangle, grid_rotation_matrix=None):
        self.grid_handler.handle_hover_over(grid=self.grid,
                                            area_rect=window_rect,
                                            on_hover_handlers=[
                                                lambda hover: self.pff.calculate_if_passed_threshold(hover)],
                                            on_hover_changed_handlers=[lambda old, new: self.pff.set_start_timer()],
                                            draw_scale_matrix=grid_rotation_matrix)

    def redraw_graph_grid_overlay(self,
                                  grid_surface: pygame.Surface,
                                  graph_surface: pygame.Surface,
                                  overlay_surface: pygame.Surface,
                                  rotation_matrix: np.ndarray,
                                  highlight_mouse_grid_pos_color: Color = Color.AQUA,
                                  predefined_highlights: Dict[Vector2, Color] = None,
                                  path_color: Color = Color.MAGENTA,
                                  path_node_color: Color = None
                                  ):

        # Resolve params into mouse pos args
        mouse_pos_args = self._resolve_mouse_highlight_args(overlay_surface,
                                                            highlight_mouse_grid_pos_color)

        # draw graph/grid
        self.draw(grid_surface=grid_surface,
                  graph_surface=graph_surface,
                  overlay_surface=overlay_surface,
                  rotation_matrix=rotation_matrix,
                  predefined_highlights=predefined_highlights,
                  highlight_mouse_pos=mouse_pos_args
                  )

        # Draw ASTAR Lines onto overlay
        self.draw_calculated_path(grid_graph=self.grid_graph,
                                  astar_results=self.pff.astar_result,
                                  path_color=path_color,
                                  window_rect=mouse_pos_args.windowRect,
                                  surface=overlay_surface,
                                  rotation_matrix=rotation_matrix,
                                  path_node_color=path_node_color)

    def redraw_graph(self,
                     graph_surface: pygame.Surface,
                     rotation_matrix: np.ndarray,
                     ):

        # draw graph
        self.draw(graph_surface=graph_surface,
                  rotation_matrix=rotation_matrix
                  )

    def redraw_grid_overlay(self,
                            overlay_surface: pygame.Surface,
                            rotation_matrix: np.ndarray,
                            highlight_mouse_grid_pos_color: Color = Color.AQUA,
                            predefined_highlights: Dict[Vector2, Color] = None,
                            path_color: Color = Color.MAGENTA,
                            path_node_color: Color = None):

        # Resolve params into mouse pos args
        mouse_pos_args = self._resolve_mouse_highlight_args(overlay_surface,
                                                            highlight_mouse_grid_pos_color)

        # draw overlay
        self.draw(overlay_surface=overlay_surface,
                  rotation_matrix=rotation_matrix,
                  predefined_highlights=predefined_highlights,
                  highlight_mouse_pos=mouse_pos_args
                  )

        # Draw ASTAR Lines onto overlay
        self.draw_calculated_path(grid_graph=self.grid_graph,
                                  astar_results=self.pff.astar_result,
                                  path_color=path_color,
                                  window_rect=mouse_pos_args.windowRect,
                                  surface=overlay_surface,
                                  rotation_matrix=rotation_matrix,
                                  path_node_color=path_node_color)

    # def redraw_hover_grid_pos_overlay(self,
    #                                   hover_grid_surface: pygame.Surface,
    #                                   highlight_mouse_grid_pos_color: Color,
    #                                   rotation_matrix: np.ndarray):
    #     # Resolve params into mouse pos args
    #     mouse_pos_args = self._resolve_mouse_highlight_args(hover_grid_surface,
    #                                                         highlight_mouse_grid_pos_color)
    #
    #     # draw overlay
    #     self.draw(hover_pos_surface=hover_grid_surface,
    #               rotation_matrix=rotation_matrix,
    #               highlight_mouse_pos=mouse_pos_args
    #               )

    def resolve_mouse_highlight_args(self,
                                      surface: pygame.Surface,
                                      highlight_mouse_grid_pos_color: pygame.Color):
        surface_rect = surface.get_rect()
        window_rect = Rectangle.from_tuple(surface_rect)
        mouse_pos_args = MouseGridResolverArgs(window_rect,
                                               help.mouse_pos_as_vector(),
                                               color=highlight_mouse_grid_pos_color) if highlight_mouse_grid_pos_color else None
        return mouse_pos_args

    def act_on_grid(self, grid_pos: Vector2, policies=List[IActOnDictPolicy]):
        state = self.grid.act_on_grid(grid_pos.y, grid_pos.x, [self.toggle_bool_policy])
        self._invalidated_highlights = True


from typing import Dict
from cooptools.colors import Color

class RuntimeConfig:
    def __init__(self, runtime_config: Dict = None):
        self.runtime_config = runtime_config if runtime_config else {}
        self._txt__fps_txt_offset_x = "fps_txt_offset_x"
        self._txt__fps_txt_offset_y = "fps_txt_offset_y"
        self._txt__fps_txt_font_size = "fps_txt_font_size"
        self._txt__mousecoords_txt_offset_x = "mousecoords_txt_offset_x"
        self._txt__mousecoords_txt_offset_y = "mousecoords_txt_offset_y"
        self._txt__mousecoord_txt_font_size = "mousecoords_txt_font_size"
        self._txt__mousecoords_color_rgb = "mousecoords_color_rgb"
        self._txt__monitoredclass_txt_offset_x = "monitoredclass_txt_offset_x"
        self._txt__monitoredclass_txt_offset_y = "monitoredclass_txt_offset_y"
        self._txt__monitoredclass_txt_font_size = "monitoredclass_txt_font_size"
        self._txt__monitoredclass_color_rgb = "monitoredclass_color_rgb"
        self._txt__debug_mode = "debug_mode"
        self._txt__fullscreen = "fullscreen"
        self._txt__screen_width = "screen_width"
        self._txt__screen_height = "screen_height"
        self._txt__max_fps = "max_fps"
        self._txt__fps_color_rgb = "fps_color_rgb"
        self._txt__log_tracked_time_interval_sec = "log_tracked_time_interval_sec"

    @property
    def log_tracked_time_interval_sec(self):
        return int(self.runtime_config.get(self._txt__log_tracked_time_interval_sec, 5))

    @property
    def fps_txt_offset_x(self):
        return int(self.runtime_config.get(self._txt__fps_txt_offset_x, 0))

    @property
    def fps_txt_offset_y(self):
        return int(self.runtime_config.get(self._txt__fps_txt_offset_y, 0))

    @property
    def fps_txt_font_size(self):
        return int(self.runtime_config.get(self._txt__fps_txt_font_size, 10))

    @property
    def fps_color(self):
        evald_rgb = eval(self.runtime_config.get(self._txt__fps_color_rgb, "(255, 255, 255)"))
        return Color.closest_color(evald_rgb)

    @property
    def mousecoords_txt_offset_x(self):
        return int(self.runtime_config.get(self._txt__mousecoords_txt_offset_x, 0))

    @property
    def mousecoords_txt_offset_y(self):
        return int(self.runtime_config.get(self._txt__mousecoords_txt_offset_y, 13))

    @property
    def mousecoords_txt_font_size(self):
        return int(self.runtime_config.get(self._txt__mousecoord_txt_font_size, 10))

    @property
    def mousecoords_txt_color(self):
        evald_rgb = eval(self.runtime_config.get(self._txt__mousecoords_color_rgb, "(255, 255, 255)"))
        return Color.closest_color(evald_rgb)

    @property
    def monitoredclass_txt_offset_x(self):
        return int(self.runtime_config.get(self._txt__monitoredclass_txt_offset_x, 0))

    @property
    def monitoredclass_txt_offset_y(self):
        return int(self.runtime_config.get(self._txt__monitoredclass_txt_offset_y, 0))

    @property
    def monitoredclass_txt_font_size(self):
        return int(self.runtime_config.get(self._txt__monitoredclass_txt_font_size, 10))

    @property
    def monitoredclass_txt_color(self):
        evald_rgb = eval(self.runtime_config.get(self._txt__monitoredclass_color_rgb, "(255, 255, 255)"))
        return Color.closest_color(evald_rgb)

    @property
    def debug_mode(self):
        return bool(self.runtime_config.get(self._txt__debug_mode, False))

    @property
    def fullscreen(self):
        return bool(self.runtime_config.get(self._txt__fullscreen, False))

    @property
    def screen_dims(self):
        width = int(self.runtime_config.get(self._txt__screen_width, 1000))
        height = int(self.runtime_config.get(self._txt__screen_height, 1000))
        return (width, height)

    @property
    def max_fps(self):
        return int(self.runtime_config.get(self._txt__max_fps, 120))


from dataclasses import dataclass, field, asdict, fields, MISSING
from typing import Dict, Iterable, List
from cooptools.common import verify_val, bound_value, verify
import uuid
from cooptools.commandDesignPattern import CommandController, CommandProtocol
import datetime
import cooptools.date_utils as du
import cooptools.config as conf
import cooptools.os_manip as osm
from cooptools.typeProviders import StringProvider, resolve_string_provider
from coopgame.characterDevelopment.manifests import MutuallyExclusiveManifest

@dataclass(frozen=True, slots=True)
class StatMeta:
    min: float = None
    max: float = None

@dataclass(frozen=True, slots=True)
class StatState:
    stat_name: str
    val: float
    description: str = field(default="")
    meta: StatMeta = None

    def with_(self,
            stat_name: str = None,
            val: float = None,
            meta: StatMeta = None
    ):
        obj_dict = asdict(self)

        obj_dict.update({k: v for k, v in {
            f'{self.stat_name=}'.split('=')[0].replace('self.', ''): stat_name,
            f'{self.val=}'.split('=')[0].replace('self.', ''): val,
            f'{self.meta=}'.split('=')[0].replace('self.', ''): meta,
        }.items() if v is not None})

        return StatState(**obj_dict)

    def __post_init__(self):
        if type(self.meta) == dict:
            object.__setattr__(self, f'{self.meta=}'.split('=')[0].replace('self.', ''), StatMeta(**self.meta))

        verify_val(self.val, gte=self.meta.min, lte=self.meta.max)



@dataclass(frozen=True, slots=True)
class StatEffect:
    stat_name: str
    new_name: str = None
    set_effect: float = None
    delta_effect: float = None
    new_meta: StatMeta = None

    @classmethod
    def from_delta(cls, name: str, delta: float):
        return StatEffect(stat_name=name, delta_effect=delta)

    def with_(self,
        stat_name: str=None,
        new_name: str = None,
        set_effect: float = None,
        delta_effect: float = None,
        new_meta: StatMeta = None,
    ):
        obj_dict = asdict(self)

        obj_dict.update({k: v for k, v in {
            f'{self.stat_name=}'.split('=')[0].replace('self.', ''): stat_name,
            f'{self.new_name=}'.split('=')[0].replace('self.', ''): new_name,
            f'{self.set_effect=}'.split('=')[0].replace('self.', ''): set_effect,
            f'{self.delta_effect=}'.split('=')[0].replace('self.', ''): delta_effect,
            f'{self.new_meta=}'.split('=')[0].replace('self.', ''): new_meta,
        }.items() if v is not None})

        return StatEffect(**obj_dict)

    def opposite(self):
        return self.with_(
            delta_effect=-self.delta_effect
        )

    def apply(self, current_state: StatState) -> StatState:
        if not verify(lambda: self.stat_name==current_state.stat_name, msg=f"Cannot apply effect to stat: {self.stat_name} != {current_state.stat_name}", block=False):
            return current_state

        new = self.set_effect if self.set_effect is not None else current_state.val
        new += self.delta_effect if self.delta_effect is not None else 0

        return current_state.with_(
            stat_name=self.new_name,
            val=bound_value(val=new, gte=current_state.meta.min, lte=current_state.meta.max) if any([self.set_effect, self.delta_effect]) else None,
            meta=self.new_meta
        )

    def summary(self)-> Dict:
        raw = asdict(self)
        raw[f'{self.new_meta=}'.split('=')[0].replace('self.', '')] = asdict(self.new_meta) if self.new_meta else None

        return {k: v for k, v in raw.items() if v is not None}

@dataclass(frozen=True, slots=True)
class StatStates:
    stats: Dict[str, StatState] = field(default_factory=dict)

    def init_new_stat(self,
                  state: StatState):
        new_stats = self.stats
        new_stats[state.stat_name] = state

        return StatStates(
            stats=new_stats
        )

    def with_(self,
              stat_effects: Iterable[StatEffect]=None):
        if stat_effects is None:
            return self

        new_stats = self.stats
        for effect in stat_effects:
            if effect.stat_name not in new_stats.keys() and effect.set_effect is None and effect.new_meta is None:
                raise ValueError(f"Cannot set a new stat {effect} without a default value for effect.set_effect and effect.new_meta")
            if effect.stat_name not in new_stats.keys():
                self.init_new_stat(state=StatState(
                    stat_name=effect.stat_name,
                    val=effect.set_effect,
                    meta=effect.new_meta
                ))

            new_stats[effect.stat_name] = effect.apply(self.stats[effect.stat_name])


        if len(new_stats) == 0:
            deb = True
        return StatStates(
            stats=new_stats
        )

    @property
    def Stats(self)-> Dict[str, StatState]:
        return self.stats

    @property
    def Summary(self)-> Dict[str, float]:
        return {
            k: v.val for k, v in self.stats.items()
        }

@dataclass(frozen=True, slots=True)
class StatEffectPackage:
    stat_effects: frozenset[StatEffect]
    description: str = field(default="")
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    category: str = field(default='DEFAULT')
    stat_opinions: Dict[str, float] = field(default_factory=dict)

    def __post_init__(self):
        if type(self.stat_effects) == dict:
            object.__setattr__(self, f'{self.stat_effects=}'.split('=')[0].replace('self.', ''), frozenset([StatEffect(**x) for x in self.stat_effects.values()]))

        if self.stat_opinions is None:
            object.__setattr__(self, f'{self.stat_opinions=}'.split('=')[0].replace('self.', ''), {})

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return hash(other) == hash(self)

    def summary(self) -> Dict:
        raw = asdict(self)

        raw[f'{self.stat_effects=}'.split('=')[0].replace('self.', '')] = [x.summary() for x in self.stat_effects]

        return {k: v for k, v in raw.items() if v is not None}

    def with_opposite_stat_effects(self, id: str = None, description: str = None):
        new = [x.opposite() for x in self.stat_effects]

        if description is None: description = f"[OPPOSITE OF] {self.description}"
        return StatEffectPackage(
            stat_effects=frozenset(new),
            description=description,
            category=self.category,
            id=id
        )

class StatEffectPackageManifest:
    def __init__(self, manifests: Dict[str, MutuallyExclusiveManifest] = None):
        self.stat_effect_type_manifests: Dict[str, MutuallyExclusiveManifest] = manifests if manifests is not None else {}

    @classmethod
    def from_config_data(self, config_file_name_provider: StringProvider):
        config = conf.JsonConfigHandler(file_path_provider=config_file_name_provider,
                                        file_creation_args=conf.FileCreationArgs(fileType=osm.FileType.JSON))
        traits_types: Dict[str, Dict] = config.resolve('traits')
        trait_details: Dict[str, Dict] = config.resolve('trait_details')

        traits = {
            k: StatEffectPackage(stat_effects=frozenset([StatEffect(stat_name=name,
                                                                   delta_effect=val) for name, val in v.get('stat_effects', {}).items()]),
                                 id=k,
                                 category=next(type for type, options in traits_types.items()
                                               if any(k in option_lst for option_lst in options.values())),
                                 stat_opinions=v.get('trait_opinions', None)
                                 )
            for k, v in trait_details.items()
        }

        stat_effect_type_manifests = {
            trait_type: MutuallyExclusiveManifest(
                options={
                    k: [traits[v] for v in options] for k, options in trait_relationship.items()
                }) for trait_type, trait_relationship in traits_types.items()
        }
        return StatEffectPackageManifest(
            manifests=stat_effect_type_manifests
        )

@dataclass(frozen=True, slots=True)
class StatsPackageState:
    base_stats: StatStates = field(default_factory=StatStates)
    stat_modifiers: frozenset[StatEffectPackage] = field(default_factory=frozenset)

    def with_(self,
              base_stat_effect_modifications: StatEffectPackage=None,
              added_modifiers: Iterable[StatEffectPackage] = None,
              removed_modifiers: Iterable[StatEffectPackage] = None,
              ):

        # update skills set
        new_skills = set(self.stat_modifiers)
        new_skills.update(added_modifiers)
        new_skills = [x for x in new_skills if x not in removed_modifiers]

        base_stat_effects = base_stat_effect_modifications.stat_effects if base_stat_effect_modifications is not None else None

        return StatsPackageState(
            base_stats=self.base_stats.with_(
                stat_effects=base_stat_effects
            ),
            stat_modifiers=frozenset(new_skills)
        )

    def resolve_stat(self,
                     stat_name: str,
                     external_stat_modifiers: Iterable[StatEffectPackage] = None) -> StatState:
        stat = self.base_stats.Stats[stat_name]

        all_mods = list(self.stat_modifiers) + (list(external_stat_modifiers) if external_stat_modifiers is not None else [])

        for mod in all_mods:
            for stat_effect in mod.stat_effects:
                stat = stat_effect.apply(stat)

        return stat

    @classmethod
    def from_config(self, config_file_name_provider: StringProvider):
        config = conf.JsonConfigHandler(file_path_provider=config_file_name_provider,
                                   file_creation_args=conf.FileCreationArgs(fileType=osm.FileType.JSON))

        stats: Dict[str, Dict] = config.resolve('stats')

        return StatsPackageState(
            base_stats=StatStates(stats={k: StatState(stat_name=k,
                                                      val=v['base_val'],
                                                      meta=StatMeta(min=v['min_val'] if 'min_val' in v.keys() else None,
                                                                    max=v['max_val'] if 'max_val' in v.keys() else None),
                                                      description=v["description"]) for k, v in stats.items()})
        )



    @property
    def InternallyResolvedStats(self) -> StatStates:
        return StatStates(
            stats={x: self.resolve_stat(
                x) for x in self.BaseStats.stats.keys()}
        )

    @property
    def BaseStats(self) -> StatStates:
        return self.base_stats

    @property
    def StatModifiers(self) -> List[StatEffectPackage]:
        return list(self.stat_modifiers)

    @property
    def Summary(self) -> Dict:
        return {
            'base_stats': self.BaseStats.Summary,
            'stat_modifiers': [x.summary() for x in self.stat_modifiers],
            'internally_resolved_stats': self.InternallyResolvedStats.Summary,
        }


class StatsPackageStateEvent(CommandProtocol):
    def __init__(self,
                 description: str,
                 base_stat_modifiers: StatEffectPackage=None,
                 added_stat_modifiers: Iterable[StatEffectPackage]=None,
                 removed_stat_modifiers: Iterable[StatEffectPackage]=None,
                 date: datetime.datetime = None,
                 id: str = None
                 ):
       self.description: str = description
       self.base_stat_modifiers: StatEffectPackage = base_stat_modifiers
       self.added_stat_modifiers: List[StatEffectPackage] = list(added_stat_modifiers) if added_stat_modifiers is not None else []
       self.removed_stat_modifiers: List[StatEffectPackage] = list(removed_stat_modifiers) if removed_stat_modifiers is not None else []
       self.date: datetime.datetime = date if date is not None else du.today(remove_ms=True)
       self.id: str = id if id is not None else str(uuid.uuid4())

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __repr__(self):
        return f"""{self.date} - {self.description} [{self.id}]
                    \tbase_stat_modifiers: {self.base_stat_modifiers}
                    \tadded_stat_modifiers: {self.added_stat_modifiers}
                    \tremoved_stat_modifiers: {self.removed_stat_modifiers}"""

    def execute(self, state: StatsPackageState):
        return state.with_(
            base_stat_effect_modifications=self.base_stat_modifiers,
            added_modifiers=self.added_stat_modifiers,
            removed_modifiers=self.removed_stat_modifiers
        )

class InitStatStatsPackageStateEvent(StatsPackageStateEvent):
    def __init__(self,
                 stat_args: Iterable[StatState]
                 ):
        super().__init__(
            description="Init Char Stats",
            base_stat_modifiers=StatEffectPackage(
                stat_effects=frozenset([StatEffect(v.stat_name,
                                                        set_effect=v.val,
                                                        new_meta=v.meta) for v in stat_args]))
        )

class DeltaStatStatsPackageStateEvent(StatsPackageStateEvent):
    def __init__(self,
                 stat_name: str,
                 delta: float,
                 description: str=None):
        if description is None: description = f"Adjust [{stat_name}] by {delta}"
        super().__init__(
            description=description,
            base_stat_modifiers=StatEffectPackage(
                stat_effects=frozenset([StatEffect(stat_name, delta_effect=delta)]))
            )


class StatsPackageCommandController:
    def __init__(self,
                 init_stats_package_state: StatsPackageState = None):
        self._init_state = StatsPackageState()
        self._event_command_controller: CommandController = CommandController(init_state=self._init_state, cache_interval=10)
        self.id = str(uuid.uuid4())

        if init_stats_package_state is not None:
            self.init_stats(init_stats_package_state.base_stats.stats.values())
            self.acquire_stat_effect_packages(init_stats_package_state.stat_modifiers)


    def __repr__(self):
        return str(self._event_command_controller.State)

    def register_events(self, events: Iterable[StatsPackageStateEvent]):
        self._event_command_controller.execute(commands=list(events))

    @property
    def State(self) -> StatsPackageState:
        return self._event_command_controller.State

    @property
    def Events(self) -> List[StatsPackageStateEvent]:
        return self._event_command_controller.command_store.get_commands()

    def init_stats(self,
                    stat_args: Iterable[StatState]
                   ):
        self.register_events(events=[InitStatStatsPackageStateEvent(stat_args)])
        return self

    def acquire_stat_effect_packages(self, stat_effect_packages: Iterable[StatEffectPackage]):
        self.register_events(events=[StatsPackageStateEvent(description="Add Stat Effect Package",
                                                            added_stat_modifiers=stat_effect_packages)])
        return self

    def lose_stat_effect_packages(self, stat_effect_packages: Iterable[StatEffectPackage]):
        self.register_events(events=[StatsPackageStateEvent(description="Add Traits",
                                                            removed_stat_modifiers=stat_effect_packages)])
        return self

if __name__ == "__main__":
    from pprint import pprint
    import coopgame.characterDevelopment.race_manifest as rm
    import coopgame.characterDevelopment.stat_modifier_manifest as smm

    def test1():
        spcc = StatsPackageCommandController()\
            .init_stats([
                StatState('hp', val=1000, meta=StatMeta(0, 1000)),
                StatState('mana', val=100, meta=StatMeta(0, 100)),
                StatState('rage', val=0, meta=StatMeta(0, 100))
            ])


        pprint(spcc.State.Summary)

        spcc.register_events(events=[
            DeltaStatStatsPackageStateEvent(stat_name='hp', delta=-19),
            StatsPackageStateEvent(description="character gained rage",
                                        added_stat_modifiers=[
                                StatEffectPackage(stat_effects=frozenset([
                                   StatEffect(stat_name='rage', delta_effect=5)]))
                            ]
                                        )
        ])

        pprint(spcc.State.Summary)
        print(f"All My Events:")
        pprint(spcc.Events)

    def test2():
        config_fn_provider = r'C:\Users\tburns\PycharmProjects\coopgame\coopgame\characterDevelopment\manifest.json'
        manif = StatEffectPackageManifest.from_config_data(config_file_name_provider=config_fn_provider)
        used_types = []

        for trait_type, trait_manif in manif.stat_effect_type_manifests.items():
            for i in trait_manif.options:
                type, choice = trait_manif.random(excluded_types=used_types)
                used_types.append(type)
                print(type, choice.id if choice else None)

    # test1()
    test2()



from coopgame.particleSystem.destroyers import particle_destroyer
from coopgame.particleSystem.callables import vec2_callable, float_from_ms_callable
from coopstructs.geometry.vectors.vectorN import Vector2
from typing import List, Union
from coopgame.particleSystem.rendering.particleRenderer import ParticleRenderer
from coopgame.particleSystem.particleGenerationArgs import ParticleGenerationArgs
from dataclasses import dataclass
from functools import partial
from coopgame.particleSystem.destroyers import time_destroyer
from coopgame.particleSystem.rendering.particleRenderer import RenderShape
from cooptools.colors import Color

@dataclass(frozen=True)
class EmitterArgs:
    origin: Union[Vector2, vec2_callable]
    velocity: Union[Vector2, vec2_callable]
    destroyers: List[particle_destroyer]
    size: Union[float, float_from_ms_callable]
    renderer: ParticleRenderer
    accel: Union[Vector2, vec2_callable] = None
    particle_generation_args: ParticleGenerationArgs = None

    def __post_init__(self):
        if self.destroyers is None or len(self.destroyers) == 0:
            raise ValueError(f"At least one particle destroyer must be provided")

        if self.particle_generation_args is None:
            object.__setattr__(self, 'particle_generation_args', ParticleGenerationArgs())

        if self.renderer is None:
            object.__setattr__(self, 'renderer', ParticleRenderer(
            shape=RenderShape.CIRCLE,
            colorBehavior=Color.SNOW
        ))

    def resolve_origin(self):
        if callable(self.origin):
            _o = self.origin()
        else:
            _o = self.origin

        return _o

    def resolve_velocity(self):
        if callable(self.velocity):
            _v = self.velocity()
        else:
            _v = self.velocity

        return _v

    def resolve_size(self, x):
        if callable(self.size):
            _s = self.size(x)
        else:
            _s = self.size

        return _s

def emitter_args_factory(
        args: EmitterArgs = None,
        origin: Union[Vector2, vec2_callable] = None,
        velocity: Union[Vector2, vec2_callable] = None,
        destroyers: List[particle_destroyer] = None,
        size: Union[float, float_from_ms_callable] = None,
        renderer: ParticleRenderer = None,
        accel: Union[Vector2, vec2_callable] = None,
        particle_generation_args: ParticleGenerationArgs = None) -> EmitterArgs:

    origin = origin or (args.origin if args else None) or Vector2(0,0)
    velocity = velocity or (args.velocity if args else None) or Vector2(0,0)
    destroyers = destroyers or (args.destroyers if args else None) or [partial(time_destroyer, lifespan_ms=2500)]
    size = size or (args.size if args else None) or 5
    renderer = renderer or (args.renderer if args else None)
    accel = accel or (args.accel if args else None) or Vector2(0, 0)
    particle_generation_args = particle_generation_args or (args.particle_generation_args if args else None)

    return EmitterArgs(
        origin=origin,
        velocity=velocity,
        destroyers=destroyers,
        size=size,
        renderer=renderer,
        accel=accel,
        particle_generation_args=particle_generation_args
    )





import time
import coopgame.pygamehelpers as help
import pygame


class GameTimeTracker:
    def __init__(self, max_fps):
        self.ticks = 0
        self.frames = 0
        self.frame_times = []
        self.fps = None
        self.max_fps = max_fps

        self.clock = pygame.time.Clock()
        self.game_start = None

    def recalculate_fps(self, ticks_last_frame: int):
        if len(self.frame_times) > 20:
            self.frame_times.pop(0)

        self.frame_times.append(ticks_last_frame)

        self.fps = help.calculate_fps(self.frame_times)
        return self.fps

    def set_start(self):
        self.game_start = time.perf_counter()

    @property
    def run_time(self):
        return (time.perf_counter() - self.game_start) if self.game_start else None

    def update(self, time_multiplier: int|float = 1.0) -> int:
        t = pygame.time.get_ticks()
        delta_time_ms = (t - self.ticks) * (time_multiplier * 1.0)
        self.ticks = t
        self.recalculate_fps(delta_time_ms)
        self.frames += 1
        self.clock.tick(self.max_fps)

        return delta_time_ms



import random as rnd
from coopgame.gameTemplate import GameTemplate
from cooptools.colors import Color
import pygame
from coopgame.renderedObjectHandling.MatrixManipulation import rotateAroundAxis
from coopstructs.geometry.vectors.vectorN import Vector2, VectorN
import numpy as np
import coopgame.handleKeyPressedArgs as hkpa
from coopgame.pygame_k_constant_names import PyKeys, PyMouse

class Particle():
    def __init__(self, id, pos, value, size, color=Color.BLUE):
        self.pos = pos
        self.size = size
        self.thickness = 1
        self.value = value
        self.color = color
        self.id = id

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def display(self, screen: pygame.Surface, color=None):
        if color is None:
            color = self.color
        pygame.draw.circle(screen, color.value, (int(self.pos[0]), int(self.pos[1])), self.size)

class Game(GameTemplate):
    def initialize_game(self):
        pass

    def handle_mouse_scroll_up(self, pressed_keys):
        pass

    def handle_mouse_scroll_down(self, pressed_keys):
        pass

    def sprite_updater(self, delta_time_ms: int):
        pass

    def handle_right_mouse_down(self, pressed_keys):
        pass

    def handle_hover_over(self, mouse_pos_as_vector: Vector2):
        pass

    def on_resize(self):
        pass

    def handle_left_mouse_down(self, pressed_keys):
        pass

    def __init__(self):
        GameTemplate.__init__(self,
                              input_state_handler=hkpa.InputStateHandler(
                                  key_actions=[
                                      hkpa.InputAction([hkpa.InputEvent(PyKeys.K_q,
                                                                        event_type=hkpa.InputEventType.DOWN)],
                                                       callback=lambda x: self._rotate_y_axis()),
                                      hkpa.InputAction([hkpa.InputEvent(PyKeys.K_a,
                                                                        event_type=hkpa.InputEventType.DOWN)],
                                                       callback=lambda x: self._rotate_x_axis())
                                  ]
                              ))
        self.particles = {}
        self.create_part_time = 0
        self.rotators = [] #np.array([10, 250, 67, 1])
        self.init_rotators()
        self.axis = np.array([1, 2, 0.5])
        self.rotation_vector = self.axis / (self.axis ** 2).sum() ** 0.5

        self.radians_per_sec = np.pi
        self.center = np.array([1000, 1000, 0])
        self.game_time = 0
        self.particle_size = 20

        self.index = -1



    def _rotate_x_axis(self):
        self.axis += np.array([.5, 0, 0])
        self.rotation_vector = self.axis / (self.axis ** 2).sum() ** 0.5


    def _rotate_y_axis(self):
        self.axis += np.array([0, 0.5, 0])
        self.rotation_vector = self.axis / (self.axis ** 2).sum() ** 0.5

    def _get_next_index(self):
        self.index += 1
        return self.index

    def init_rotators(self):
        self.rotators.append([np.array([10, 250, 67, 1]), Color.BLUE])
        self.rotators.append([np.array([250, 250, 10, 1]), Color.RED])
        self.rotators.append([np.array([750, 800, 10, 1]), Color.GREEN])


    def model_updater(self, delta_time_ms: int):
        rads = self.radians_per_sec * delta_time_ms / 1000
        matrix_other = rotateAroundAxis(self.center, self.rotation_vector, rads)

        if self.create_part_time > 100:
            self.create_part_time = 0
            for rotator in self.rotators:
                self.particles[Particle(self._get_next_index(), rotator[0], self.particle_size, 1,
                                        color=rotator[1])] = self.game_time

        for rotator in self.rotators:
            rotator[0] = np.dot(matrix_other, rotator[0])
            self.create_part_time += delta_time_ms
            self.game_time += delta_time_ms


        to_delete = []
        delete_time = 30000
        shrink_time = 3000

        for particle, create_time in self.particles.items():
            if self.game_time - create_time > delete_time:
                to_delete.append(particle)

            particle.size = int(max((1 - (self.game_time - create_time) / shrink_time) * self.particle_size, 1))

        for particle in to_delete:
            del self.particles[particle]

    def draw(self, frames:int, debug_mode:bool = False):
        for particle in self.particles.keys():
            particle.display(self.screen)

        # self.draw_fps(self.screen)

game = Game()
game.main()


rnd.seed(0)

axis = np.array([1, 0, 1])
rotation_vector = axis / (axis**2).sum()**0.5


for ii in range(0, 100):
    other = np.array([rnd.randint(0, 100), rnd.randint(0, 100), rnd.randint(0, 100)], dtype=float)
    center = np.array([rnd.randint(0, 100), rnd.randint(0, 100), rnd.randint(0, 100)], dtype=float)
    matrix_center = rotateAroundAxis(center, rotation_vector, .1)
    matrix_other = rotateAroundAxis(center, rotation_vector, np.pi * 2)

    translated_center = matrix_center.dot(np.append(center, 1))
    translated_other = matrix_other.dot(np.append(other, 1))
    # verify = translated_point[:2] == center
    if not np.allclose(translated_center[:3], center):
        print (translated_center[:3], center)
        raise Exception(f"Translated center does not equal center")

    if not np.allclose(translated_other[:3], other):
        print (matrix_other, translated_other[:3], other)
        raise Exception(f"Translated other does not equal other")



from dataclasses import dataclass, field
from typing import List
from examples.voroni_world.models import Region, Villager

@dataclass
class BehaviorResults:
    villager: Villager
    revealed_regions: List[Region] = field(default_factory=list)


import pygame

from coopgame.particleSystem.particle import Particle
from coopstructs.geometry.vectors.vectorN import Vector2
from typing import List, Dict
from coopgame.particleSystem.emitterArgs import EmitterArgs, emitter_args_factory
import uuid
from cooptools.toggles import BooleanToggleable
from coopgame.particleSystem.emitterPhaseController import EmitterPhaseController

class Emitter:

    def __init__(self,
                 args: EmitterArgs,
                 trail_args: EmitterArgs = None,
                 id = None,
                 context = None,
                 start_on_init: bool = True
                 ):

        self.particles: List[Particle] = []
        self.args = args
        self.trail_args = trail_args

        self.trails: Dict[Particle, Emitter] = {}

        self._last_creation = 0
        self._next_burst = 0
        self._now = 0
        self._n = 0
        self._n_bursts = 0

        self._lifetime_ms = 0

        self._started = BooleanToggleable(default=start_on_init)
        self.id = id or uuid.uuid4()
        self.context = context

        self._phase_controllers: List[EmitterPhaseController] = []
        self._n_phasers = 0
        self._next_phaser = 0

        self._last_pos = None

        if self.args.particle_generation_args.init_particles is not None:
            self.add_new_particles(self.args.particle_generation_args.init_particles)

    def _update_particles(self, delta_ms):
        to_remove = []
        for particle in self.particles:
            # update the particle
            particle.update(delta_ms)

            # check if particle needs destroyed
            if any(x(particle) for x in self.args.destroyers):
                to_remove.append(particle)

        # remove the particles need to be removed
        for particle in to_remove:
            self.particles.remove(particle)

    def _update_trails(self, delta_ms):
        for part, trail in self.trails.items():
            trail.update(delta_ms)

        # remove empty trails
        trails_to_remove = []
        for part, trail in self.trails.items():
            if part not in self.particles and trail.Running:
                trail.toggle_started(False)

            if part not in self.particles and trail.N_Particles == 0:
                trails_to_remove.append(part)


        for part in trails_to_remove:
            del self.trails[part]

    def _update_phases(self, delta_ms):
        to_remove = []
        for phase_controller in self._phase_controllers:
            phase_controller.update(delta_ms)
            if phase_controller.Expired:
                to_remove.append(phase_controller)

        for phase_controller in to_remove:
            self._phase_controllers.remove(phase_controller)

    def toggle_started(self, value: bool = None):
        if value:
            self._started.set_value(value)
        else:
            self._started.toggle()

    def emit(self):
        # dont handle bc passed max particles
        if self.args.particle_generation_args.MaxParticles is not None and self._n >= self.args.particle_generation_args.MaxParticles:
            return

        if self.args.particle_generation_args.burst_args is not None:
            self.handle_bursts()
        elif self.args.particle_generation_args.phase_args is not None:
            self.handle_phases()
        else:
            self.handle_stream()


    def update(self, delta_ms: int):
        self._now += delta_ms
        self._lifetime_ms += delta_ms

        if self._started.value:
            self.emit()
        else:
            self._last_creation = self._now

        # update particles
        self._update_particles(delta_ms)

        # update trails
        self._update_trails(delta_ms)

        # update phases
        self._update_phases(delta_ms)

        # store last pos
        if len(self.Particles) > 0:
            self._last_pos = self.Particles[-1].pos

    def create_trail(self, particle: Particle):
        trail = Emitter(emitter_args_factory(args=self.trail_args, origin=lambda: particle.pos), id='tail')
        self.trails[particle] = trail

    def add_new_particles(self, n_to_create, origin: Vector2 = None):
        for n in range(n_to_create):
            if origin is not None:
                _o = origin
            else:
                _o = self.args.resolve_origin()

            # create particle
            if self.args.particle_generation_args.MaxParticles is not None and self._n >= self.args.particle_generation_args.MaxParticles:
                return

            new_part = Particle(
                pos=_o,
                velo_u_s=self.args.resolve_velocity(),
                accel_u_s2=self.args.accel,
                size=self.args.resolve_size(self._lifetime_ms),
                id=f"{self.context}_{self._n}" if self.context else None
            )
            self.particles.append(new_part)

            # create a trail for the particle
            if self.trail_args:
                self.create_trail(new_part)

            self._n += 1
            self._last_creation = self._now

    def handle_phases(self):
        # dont hanlde bc passed max phasers
        if self.args.particle_generation_args.phase_args.max_phasers is not None and self._n_phasers >= self.args.particle_generation_args.phase_args.max_phasers:
            return

        # dont handle bc invterval hasnt passed
        if self._now < self._next_phaser:
            return

        # determine next phaser
        self._next_phaser = self.args.particle_generation_args.GenerationIntervalMS + self._now

        self._n_phasers += 1
        self._phase_controllers.append(EmitterPhaseController(
            self.args.particle_generation_args.phase_args
        ))


    def handle_bursts(self):
        # dont hanlde bc passed max bursts
        if self.args.particle_generation_args.burst_args.MaxBursts is not None and self._n_bursts >= self.args.particle_generation_args.burst_args.MaxBursts:
            return

        # dont handle bc invterval hasnt passed
        if self._now < self._next_burst:
            return

        self._next_burst = self.args.particle_generation_args.burst_args.BurstIntervalMS + self._now

        # determine n to create
        n_to_burst = self.args.particle_generation_args.burst_args.ParticlesPerBurst

        # add new
        self._n_bursts += 1
        self.add_new_particles(n_to_burst, origin=self.args.resolve_origin())

    def handle_stream(self):
        # dont handle bc invterval hasnt passed
        if self._now - self._last_creation < self.args.particle_generation_args.GenerationIntervalMS:
            return

        # determine n to create
        n_to_create = int((self._now - self._last_creation) / self.args.particle_generation_args.GenerationIntervalMS)

        # add new
        self.add_new_particles(n_to_create)


    def render(self, surf: pygame.Surface):
        # render my particles
        if self.args.renderer.reverse:
            for particle in reversed(self.particles):
                self.args.renderer.render(surf, particle, self)
        else:
            for particle in self.particles:
                self.args.renderer.render(surf, particle, self)

        # render my trails
        for part, trail in self.trails.items():
            trail.render(surf)

        # render phases
        for phaser in self._phase_controllers:
            phaser.render(surf)

    def n_particles(self):
        return len(self.particles) + \
               sum(trail.N_Particles for _, trail in self.trails.items()) + \
               sum(phaser.N_Particles for phaser in self._phase_controllers)

    @property
    def N_Particles(self):
        return self.n_particles()

    @property
    def Running(self):
        return self._started.value

    @property
    def Particles(self):
        return self.particles

    @property
    def LifetimeMS(self):
        return self._lifetime_ms

    @property
    def LastPos(self):
        return self._last_pos


from coopgame.gameTemplate import GameTemplate
from coopgame.pyZoneHandler import PyZoneHandler
import pygame
import coopgame.pygamehelpers as help
from coopstructs.toggles import BooleanToggleable
from coopstructs.vectors import Vector2

DUMMY_NEXT_ZONE = "DUMMY"

class ZoneDrawerGame(GameTemplate):

    def __init__(self):
        super().__init__()

        self.zone_handler = PyZoneHandler()
        self.draw_label_toggle = BooleanToggleable()

        self._zn_idx = -1

    def register_keys(self):
        # CREATE NEW ZONE
        self.register_action_to_keys((pygame.K_n,),
                                     lambda: self.handle_n(),
                                     react_while_holding=False)

        # TOGGLE ZONE LAYERS
        self.register_action_to_keys((pygame.K_UP,),
                                     lambda: self.zone_handler.layer_selector.visibility_toggle(),
                                     react_while_holding=False)

        # LABELS
        self.register_action_to_keys((pygame.K_l,),
                                     lambda: self.draw_label_toggle.toggle(),
                                     react_while_holding=False)

    def get_next_id(self):
        self._zn_idx += 1
        return str(self._zn_idx)

    def draw(self, frames:int, debug_mode: bool = False):
        self.zone_handler.draw_zones(self.screen,
                                     draw_zone_label=self.draw_label_toggle.value
                                     )

    def init_dummy_zone(self):
        if DUMMY_NEXT_ZONE in self.zone_handler.zm.zones:
            self.zone_handler.zm.delete_zone(DUMMY_NEXT_ZONE)
        self.zone_handler.init_new_zone(lambda:  DUMMY_NEXT_ZONE,
                                        initial_points=self.zone_handler.zm.zones[str(self._zn_idx)].boundary_points,
                                        set_active=False)

    def handle_n(self):
        self.zone_handler.init_new_zone(lambda: self.get_next_id())
        self.init_dummy_zone()

    def handle_left_mouse_down(self, pressed_keys):
        # set origin at mouse down point
        self.zone_handler.origin = help.mouse_pos_as_vector()


    def handle_left_mouse_up(self, pressed_keys):
        # make sure we have started a zone
        if self._zn_idx < 0: return

        # get mouse position
        mouse_pos = help.mouse_pos_as_vector()

        # determine if next point is to be placed between points based on location of mouse down (origin)
        bnds = self.zone_handler.zm.zones[str(self._zn_idx)].insersection_boundary_idxs_from_origin(
            mouse_pos, self.zone_handler.origin)

        at_idx = bnds[0][1] if bnds and len(bnds) > 0 else None
        print(bnds, at_idx)
        # add the new point
        self.zone_handler.add_to_active_zone(mouse_pos, at_idx=at_idx)

        # clear the origin
        self.zone_handler.origin = None

        # update dummy
        self.overwrite_dummy_zone_with_last_idx_zone(mouse_pos)

    def handle_right_mouse_up(self, pressed_keys):
        # get mouse position
        mouse_pos = help.mouse_pos_as_vector()

        self.zone_handler.remove_from_active_zone()
        self.overwrite_dummy_zone_with_last_idx_zone()

    def handle_hover_over(self, mouse_pos_as_vector: Vector2):
        if DUMMY_NEXT_ZONE in self.zone_handler.zm.zones \
                and self.zone_handler.origin \
                and mouse_pos_as_vector != self.zone_handler.origin:
            print(f"dragging")

            self.overwrite_dummy_zone_with_last_idx_zone(mouse_pos_as_vector)

    def overwrite_dummy_zone_with_last_idx_zone(self, mouse_pos_as_vector: Vector2 = None):
        bnds = self.zone_handler.zm.zones[str(self._zn_idx)].insersection_boundary_idxs_from_origin(
            mouse_pos_as_vector, self.zone_handler.origin)
        print(bnds)
        if mouse_pos_as_vector:
            self.zone_handler.zm.zones[DUMMY_NEXT_ZONE] = self.zone_handler.zm.zones[
                str(self._zn_idx)].with_additional_point(mouse_pos_as_vector,
                                                         at_idx=bnds[0][1] if bnds and len(bnds) > 0 else None)
            print(self.zone_handler.zm.zones[DUMMY_NEXT_ZONE])
            print(self.zone_handler.zm.zones[str(self._zn_idx)])
        else:
            self.zone_handler.zm.zones[DUMMY_NEXT_ZONE] = self.zone_handler.zm.zones[str(self._zn_idx)]



if __name__ == "__main__":

    game = ZoneDrawerGame()
    game.main()


from coopgame.grids_graphs.pygraphhandler import *
from coopgame.grids_graphs.pygridgraphHandler import *
from coopgame.grids_graphs.pygridhandler import *
from coopgame.grids_graphs.draw_grid_utils import *
from coopgame.grids_graphs.draw_graph_utils import *


from cooptools.transform import Transform
from typing import Iterable
import cooptools.geometry_utils.vector_utils as vec
import cooptools.matrixManipulation as mm
from coopstructs.geometry.rectangles.rectangle import Rectangle
import math
from cooptools.coopEnum import CardinalPosition

class Camera3d:

    def __init__(self):
        self._transform = Transform()
        self._field_of_view_rads = math.pi / 2
        self._near_plane_dist = 0.01
        self._far_plane_dist = 100

    def get_viewport_rect(self,
                          viewport_dims: vec.FloatVec) -> Rectangle:
        cx, cy = self._transform.Translation.Vector[:2]

        ox = cx - viewport_dims[0] / 2
        oy = cy - viewport_dims[1] / 2

        return Rectangle.from_tuple((ox, oy, viewport_dims[0], viewport_dims[1]))

    def scroll(self, delta_vector):
        self._transform.Translation.update(delta_vector=delta_vector)


    def zoom(self,
             rads_scalar: float = None):

        self._field_of_view_rads *= rads_scalar

    #TODO: Figure out the calc for field of view given the input rect
    # def zoom_to_rect(self,
    #                  pos: Iterable[float],
    #                  dims: Iterable[float],
    #                  cardinality: CardinalPosition = None
    #                  ):
    #
    #     if cardinality is None:
    #         cardinality = CardinalPosition.BOTTOM_LEFT
    #
    #     center = CardinalPosition.alignment_conversion(
    #         dims=dims,
    #         anchor=pos,
    #         from_cardinality=cardinality,
    #         to_cardinality=CardinalPosition.CENTER
    #     )
    #
    #     self.center_at(
    #         pos=center
    #     )
    #




    def center_around_points(self,
                             points: vec.IterVec):
        minX = min(x[0] for x in points)
        maxX = max(x[0] for x in points)
        minY = min(x[1]for x in points)
        maxY = max(x[1] for x in points)
        minZ = min(x[2]for x in points)
        maxZ = max(x[2] for x in points)

        o = (minX + (maxX - minX) / 2, minY + (maxY - minY) / 2, minZ + (maxZ - minZ) / 2)

        self.center_at(pos=o)

    def center_at(self,
                  pos: vec.FloatVec
    ):
        self._transform.Translation.update(
            vector=pos
        )

    @property
    def PerspectiveMatrix(self):
        p = mm.perspective_matrix(
            near_plane_dist=self._near_plane_dist,
            far_plane_dist=self._far_plane_dist,
            field_of_view_rads=self._field_of_view_rads
        )
        return p

    @property
    def TransformMatrix(self):
        t = self._transform.transform_matrix(
            inversions=['x', 'y', 'z']
        )
        return t

    def get_screen_adjustment_matrix(self,viewport_dims: vec.FloatVec):
        return mm.translationMatrix(viewport_dims[0] / 2, viewport_dims[1] / 2, 0)

    def view_points(self,
                    points: vec.IterVec,
                    viewport_dims: vec.FloatVec) -> vec.LstVec:
        transformed = mm.point_transform_3d(
            points=points,
            translationM=self._transform.Translation.inversion(inversion_idxs=[0, 1, 2]).TranslationMatrix,
            scaleM=self._transform.Scale.ScaleMatrix,
            rotationM=self._transform.Rotation.RotationMatrix,
            perspectiveM=self.PerspectiveMatrix,
            post_perspective_translationM=self.get_screen_adjustment_matrix(viewport_dims),
        )
        return transformed



    def viewing_matrix(self,
                       viewport_dims: vec.FloatVec,
                       swaps: Iterable[str] = None,
                       near_plane_dist: float = None,
                       far_plane_dist: float = None):

        t = self._transform.transform_matrix(
            swaps=swaps,
            inversions=['x', 'y', 'z'],
        )

        screen_center_adj = mm.translationMatrix(viewport_dims[0] / 2, viewport_dims[1] / 2, 0)

        p = mm.perspective_matrix(
            near_plane_dist=near_plane_dist,
            far_plane_dist=far_plane_dist,
            field_of_view_rads=self._field_of_view_rads
        )

        pt = p.dot(t)



        ret1 = screen_center_adj.dot(pt)
        ret = pt
        return ret


if __name__ == "__main__":
    cam = Camera3d()
    cam.center_at((500, 500))

    print(cam.get_true_pos(
        viewport_points=[(0, 0, 0), (100, 250, 0)],
        viewport_dims=(500, 250)
    ))





import logging
from examples.voroni_world.models import *
from coopgame.gameTemplate import tt
from cooptools.decor import timer
import pprint
import math
from typing import Dict, Tuple, List, Callable
from cooptools import graphs as grph
from cooptools.physics import waypointFollower as wf
from examples.voroni_world.behaviors import cmn

logger = logging.getLogger(__name__)


FARM_ACTION = "FARMING"

def _choose_farmable_region(regions: Dict[int, Region]):
    try:
        to_farm = next(v for v in regions.values()
                    if v.terrain in [TerrainType.GRASS, TerrainType.GRASSHILLS]
                    and v.visibility_level in [VisibilityLevel.REVEALED]
                    and FARM_ACTION not in v.action_states)
    except:
        to_farm = None

    return to_farm


def farm(
        villager: Villager,
        villager_current_region_provider: Callable[[Villager], Region],
        route_provider: Callable[[Region, Region], grph.AStarResults],
        route_to_waypoints_provider: Callable[[List[grph.Node]], List[wf.Waypoint]],
        village_region: Region,
        regions: Dict[int, Region],
):

    to_farm = _choose_farmable_region(
        regions=regions
    )

    if to_farm is None:
        logger.error(f"No farm-able region")
        return

    cmn.send_villager_to_region(
        villager=villager,
        region=to_farm,
        villager_current_region_provider=villager_current_region_provider,
        route_provider=route_provider,
        route_to_waypoints_provider=route_to_waypoints_provider,
    )




import pygame

from coopgame.spriteHandling.spritesheet import SpriteSheet
from coopgame.spriteHandling.spriteFolder import SpriteFolderHandler
from typing import Dict, List, Tuple
from cooptools.toggles import IntegerRangeToggleable

class SpritePoseHandler:

    @classmethod
    def from_spritesheet(cls,
                         sprite_sheet: SpriteSheet,
                         pose_def: Dict[str, Tuple[int, int, int]], #{poseName: [row, start_idx, end_idx]}
                         colorkey: int = -1, allow_wrap: bool=True):

        surf_getter = lambda row, colorkey, start, end: tuple(sprite_sheet.load_row_strip(row, colorkey=colorkey)[start:end])

        poses = {
            pose: surf_getter(attrs[0], colorkey, attrs[1], attrs[2]) for pose, attrs in pose_def.items()
        }

        return SpritePoseHandler(poses=poses)

    @classmethod
    def from_spritefolder(cls,
                          sprite_folder: SpriteFolderHandler):
        return SpritePoseHandler(sprite_folder.poses)

    def __init__(self,
                 poses: Dict[str, List[pygame.Surface]],
                 default_pose: str=None):
        self.poses = poses
        self._pose_idx = IntegerRangeToggleable(0, len(self.poses) - 1)
        self._pose_idx_map = {
            ii: name for ii, name in enumerate(self.poses.keys())
        }
        self._name_pose_idx_map = {
            name: ii for ii, name in enumerate(self.poses.keys())
        }

        self._animation_idx: IntegerRangeToggleable = self._a_valid_integer_range_for_current_pose()

        if default_pose is not None:
            self.set_pose(default_pose)

    def get_pose(self):
        return self.poses[self._pose_idx_map[self._pose_idx.value]]

    def get_current(self):
        try:
            return {
                'pose': self._pose_idx_map[self._pose_idx.value],
                'animation_idx':self._animation_idx.value,
                'surface': self.get_pose()[self._animation_idx.value]
            }

        except Exception as e:
            raise e

    def increment_pose(self, loop: bool=True):
        self._pose_idx.toggle(on_toggle_callbacks=[lambda val: self._init_animation_indexer()],
                              loop=loop)

    def decrement_pose(self, loop: bool=True):
        self._pose_idx.toggle(reverse=True,
                              on_toggle_callbacks=[lambda val: self._init_animation_indexer()],
                              loop=loop
                              )

    def set_pose(self, pose: str):
        self._pose_idx.set_value(self._name_pose_idx_map[pose],
                                 on_toggle_callbacks=[lambda val: self._init_animation_indexer()])

    def _a_valid_integer_range_for_current_pose(self):
        return IntegerRangeToggleable(0, len(self.poses[self._pose_idx_map[self._pose_idx.value]]) - 1)

    def _init_animation_indexer(self):
        self._animation_idx = self._a_valid_integer_range_for_current_pose()

    def increment_animation(self, loop: bool=True):
        self._animation_idx.toggle(loop=loop)

    def decrement_animation(self, loop: bool=True):
        self._animation_idx.toggle(reverse=True, loop=loop)

    @property
    def Shape(self) -> Tuple[int, int]:
        first_image = next(iter(self.poses.values()))[0]
        width = first_image.get_width()
        height = first_image.get_height()
        return width, height

    @property
    def AnimationEnded(self) -> bool:
        return self._animation_idx.value == self._animation_idx.max


import pygame
from typing import Tuple
from coopgame.gameTemplate import GameTemplate
import coopgame.pointdrawing.point_draw_utils as pdu
import coopgame.pygamehelpers as phelp
from cooptools.colors import Color
import logging
import coopgame.handleKeyPressedArgs as kpa
from cooptools.toggles import BooleanToggleable
import coopgame.characterDevelopment.character as char
import coopgame.characterDevelopment.stats as stat
import coopgame.label_drawing.label_drawing_utils as ldu
from coopgame.characterDevelopment.manifests import MutuallyExclusiveManifest
from coopgame.characterDevelopment.behaviorExtensions import hp as hp
from coopgame.characterDevelopment.behaviorExtensions import dual as dual

class CharDevTestGame(GameTemplate):

    def __init__(self):
        self.show_stats_toggle = BooleanToggleable(default=True)
        self.show_dual_toggle = BooleanToggleable(default=False)

        super().__init__(
            input_state_handler=self.define_input_handler()
        )
        self.char: char.Character = None
        self.stats_surf: pygame.Surface = None
        self.dual_surf: pygame.Surface = None
        self.last_char: char.Character = None

        self.reset_char()

    def reset_char(self):
        self.last_char = self.char
        self.char = self.gen_char()

        self.build_stats_surface(char=self.char)

    def build_stats_surface(self,
                            char: char.Character):
        surf = phelp.init_surface(self.screen.get_size())
        surf.fill(color=Color.BEIGE.value)
        self.stats_surf = self.print_char_stats_to_surface(char, surf=surf)

    def build_dualing_surface(self,
                              dualist1: char.Character,
                              dualist2: char.Character):
        surf = phelp.init_surface(self.screen.get_size())
        surf.fill(color=Color.ORANGE.value)
        self.print_dualing_stats_to_surface(surf, dualist1, dualist2)
        return surf


    def draw(self, frames: int, debug_mode: bool = False):
        pdu.draw_points(
            points={
                (self.ScreenCenter): pdu.DrawPointArgs(
                    color=Color.BLUE,
                    outline_color=Color.ORANGE,
                    radius=50,
                    outline_width=4
                )
            },
            surface=self.screen
        )
        phelp.draw_arrow(self.screen,
                        self.ScreenCenter,
                        phelp.mouse_pos_as_vector().as_tuple(),
                        Color.GREEN,
                         leader_line=True)
        if self.show_stats_toggle.value:
            self.screen.blit(self.stats_surf, (0, 0))

    def define_input_handler(self):
        return kpa.InputStateHandler(
            quit_callback_package=kpa.CallbackPackage(down=lambda x: self.quit()),
            debug_callback_package=kpa.CallbackPackage(down=lambda x: self.debug_mode_toggle.toggle()),
            fullscreen_callback_package=kpa.CallbackPackage(down=lambda x: self.toggle_fullscreen()),
            key_actions=[
                kpa.InputAction([kpa.InputEvent(kpa.PyKeys.K_s)], callback=lambda x: self.show_stats_toggle.toggle()),
                kpa.InputAction([kpa.InputEvent(kpa.PyKeys.K_r)], callback=lambda x: self.reset_char()),
                kpa.InputAction([kpa.InputEvent(kpa.PyKeys.K_d)], callback=lambda x: self.dual()),
        ])

    def gen_char(self) -> char.Character:
        config_fn_provider = r'C:\Users\Tj Burns\PycharmProjects\coopgame\coopgame\characterDevelopment\manifest.json'

        names = MutuallyExclusiveManifest.from_config(config_file_name_provider=config_fn_provider, key='names')

        me = char.Character.generate(
            args=char.CharacterGeneratorArgs(
                name_provider=lambda x: names.random(type=x)[1],
                init_stats_package=stat.StatsPackageState.from_config(config_fn_provider),
                traits_manifest=stat.StatEffectPackageManifest.from_config_data(config_fn_provider)
            ),
            trait_count_args={
                "personality_traits": 3,
                "physical_traits": 2,
                "race": 1,
                "skills": 1
            }
        )
        return me

    def _dual_epoch_handler(self,
                            dualist1: char.Character,
                            dualist2: char.Character,
                            ) -> Tuple[char.Character, char.Character]:
        d1, d2 = dual.basic_dualing_epoch_handler(dualist1, dualist2)



        return d1, d2

    def dual(self):



        if self.last_char is not None:
            dual.dual(self.char, self.last_char, dual.basic_dualing_epoch_handler)



    def print_char_stats_to_surface(self,
                                    character: char.Character,
                                    surf: pygame.Surface) -> pygame.Surface:
        stats = character.stats.State.Summary

        header_font = pygame.font.SysFont(name='Comic Sans MS', size=24)
        header_font.set_underline(True)
        regular_font = pygame.font.SysFont(name='Comic Sans MS', size=12)

        ldu.draw_label(
            hud=surf,
            args=ldu.LabelArgs(
                draw_args=ldu.DrawLabelArgs(
                color=Color.BLACK,
                font=header_font,
                ),
                text=f"Stats Page",
                pos=(50, 15)
            )
        )

        n_stats = len(stats['base_stats'])
        stat_per_col = n_stats // 2

        ldu.draw_label(hud=surf,
                       args=ldu.LabelArgs(
                           text=f"Character Name: {character.name}",
                        draw_args=ldu.DrawLabelArgs(
                           color=Color.BLACK,
                           font=regular_font),
                       pos=(100, 50)
                       )
                       )


        ldu.draw_label(
            hud=surf,
            args=ldu.LabelArgs(
                text=f"Race: {character.Race.id}",
                draw_args=ldu.DrawLabelArgs(
                    color=Color.BLACK,
                    font=regular_font
                ),
                pos=(100, 60))
        )


        ii = 0
        row_offset = 15
        col_offset = 250
        for stat, val in stats['base_stats'].items():
            new_txt = f"{stat}: {val} ({stats['internally_resolved_stats'][stat]})"
            pos = (
                125 if ii < stat_per_col else 125 + col_offset,
                100 + ii * row_offset if ii < stat_per_col else 100 + (ii - stat_per_col) * row_offset
            )

            ldu.draw_label(hud=surf,
                           args=ldu.LabelArgs(
                               text=new_txt,
                               draw_args=ldu.DrawLabelArgs(
                                   color=Color.BLACK,
                                   font=regular_font
                               ),
                               pos=pos)
                           )

            ii+=1

        ii = 0
        for mod in stats['stat_modifiers']:
            lamb = lambda x: f"{x['stat_name']}: {x['delta_effect']}"
            new_txt = f"{mod['id']} [{','.join([lamb(x) for x in mod['stat_effects']])}]"
            pos = (
                125,
                300 + ii * row_offset
            )

            ldu.draw_label(hud=surf,
                           args=ldu.LabelArgs(
                               text=new_txt,
                               draw_args=ldu.DrawLabelArgs(
                                   color=Color.BLACK,
                                   font=regular_font
                               ),
                               pos=pos)
                           )

            ii+=1

        if self.last_char is not None:
            opinion = self.char.opinon_of(self.last_char)
            ldu.draw_label(hud=surf,
                           args=ldu.LabelArgs(
                               text=f"Opinion of {self.last_char.name}: {opinion.Score}",
                               draw_args=ldu.DrawLabelArgs(
                                   color=Color.BLACK,
                                   font=regular_font
                               ),
                               pos=(700, 75))
                           )


            x_start = 725
            y_start = 90
            ii=0
            for stat, val in opinion.factors.items():
                new_txt = f"{stat}: {val}"
                pos = (
                    x_start,
                    y_start + ii * row_offset
                )

                ldu.draw_label(hud=surf,
                               args=ldu.LabelArgs(
                                   text=new_txt,
                                   draw_args=ldu.DrawLabelArgs(
                                       color=Color.BLACK,
                                       font=regular_font
                                   ),
                                   pos=pos)
                               )

                ii += 1

        return surf

    def _get_font(self,
                  underlined: bool = False,
                  size: int = 12,
                  name: str = 'Comic Sans MS'):
        ret = pygame.font.SysFont(name=name, size=size)
        if underlined:
            ret.set_underline(True)
        return ret


    def print_dualing_stats_to_surface(self,
                                        surf: pygame.Surface,
                                        dualist1: char.Character,
                                        dualist2:char.Character,

    ):
        header_font = self._get_font(underlined=True, size=24)
        regular_font = self._get_font()

        # draw header
        ldu.draw_label(hud=surf,
                       args=ldu.LabelArgs(
                           text=f"DUAL!!",
                           draw_args=ldu.DrawLabelArgs(
                               font=header_font,
                               color=Color.BLACK
                           ),
                           pos=(50, 15))
                       )


        # draw dualist 1
        ldu.draw_label(hud=surf,
                       args=ldu.LabelArgs(
                           text=f"Character Name: {dualist1.name}",
                           draw_args=ldu.DrawLabelArgs(
                               color=Color.BLACK,
                               font=regular_font
                           ),
                           pos=(100, 50))
                       )


        health = dualist1.stats.State.resolve_stat(stat_name=hp.HP)
        ldu.draw_label(hud=surf,
                       args=ldu.LabelArgs(
                           text=f"Hp: {health}",
                           draw_args=ldu.DrawLabelArgs(
                               color=Color.BLACK,
                               font=regular_font
                           ),
                           pos=(100, 60))
                       )


        # draw dualist 1
        ldu.draw_label(hud=surf,
                       args=ldu.LabelArgs(
                           text=f"Character Name: {dualist2.name}",
                           draw_args=ldu.DrawLabelArgs(
                               color=Color.BLACK,
                               font=regular_font
                           ),
                           pos=(600, 50))
                       )


        health = dualist2.stats.State.resolve_stat(stat_name=hp.HP)
        ldu.draw_label(hud=surf,
                       args=ldu.LabelArgs(
                           text=f"Hp: {health}",
                           draw_args=ldu.DrawLabelArgs(
                               color=Color.BLACK,
                               font=regular_font
                           ),
                           pos=(600, 60))

                       )

    # def print_details_to_surface(self,
    #                             surf: pygame.Surface,
    #                             text: str,
    #                             draw_label_args: ldu.DrawLabelArgs):



logging.basicConfig(level=logging.INFO)
logging.getLogger('cooptools.CommandController').setLevel(logging.WARN)
g = CharDevTestGame()
g.main()


from cooptools.geometry_utils import vector_utils as vec
from coopgame.linedrawing import line_draw_utils as line
from shapely.geometry import Polygon
from cooptools.coopEnum import CoopEnum, auto
from dataclasses import dataclass, field
from cooptools.coopDataclass import BaseIdentifiedDataClass, UniqueIdentifier
from typing import List, Dict, Set, Tuple, Self, Iterator
import logging
from cooptools.colors import Color
import uuid

logger = logging.getLogger(__name__)


class TerrainType(CoopEnum):
    DEEPSEAWATER = auto()
    SHALLOWSEAWATER = auto()
    BEACH = auto()
    GRASS = auto()
    GRASSHILLS = auto()
    FOREST = auto()
    PLAINS = auto()
    PLAINSHILLS=auto()
    MOUNTAINS = auto()
    SNOWPEAK = auto()
    UNDEFINED = -1

TERRAIN_COMPATIBLITY_MAP = {
    TerrainType.DEEPSEAWATER: [TerrainType.DEEPSEAWATER, TerrainType.SHALLOWSEAWATER],
    TerrainType.SHALLOWSEAWATER: [TerrainType.DEEPSEAWATER, TerrainType.SHALLOWSEAWATER, TerrainType.BEACH],
    TerrainType.BEACH: [TerrainType.SHALLOWSEAWATER, TerrainType.BEACH, TerrainType.PLAINS, TerrainType.GRASS],
    TerrainType.GRASS: [TerrainType.FOREST, TerrainType.GRASS, TerrainType.GRASSHILLS, TerrainType.PLAINS, TerrainType.PLAINSHILLS, TerrainType.BEACH],
    TerrainType.GRASSHILLS: [TerrainType.GRASS, TerrainType.GRASSHILLS, TerrainType.PLAINS, TerrainType.PLAINSHILLS, TerrainType.MOUNTAINS],
    TerrainType.PLAINS: [TerrainType.GRASS, TerrainType.GRASSHILLS, TerrainType.PLAINS, TerrainType.PLAINSHILLS, TerrainType.BEACH],
    TerrainType.PLAINSHILLS: [TerrainType.GRASS, TerrainType.GRASSHILLS, TerrainType.PLAINS, TerrainType.PLAINSHILLS, TerrainType.MOUNTAINS],
    TerrainType.MOUNTAINS: [TerrainType.PLAINSHILLS, TerrainType.GRASSHILLS, TerrainType.MOUNTAINS, TerrainType.SNOWPEAK],
    TerrainType.SNOWPEAK: [TerrainType.MOUNTAINS, TerrainType.SNOWPEAK],
    TerrainType.FOREST: [TerrainType.GRASS, TerrainType.FOREST],
    TerrainType.UNDEFINED: []
}

TERRAIN_COLOR_MAP = {
    TerrainType.DEEPSEAWATER: Color.REGAL_NAVY,
    TerrainType.SHALLOWSEAWATER: Color.FRESH_SKY,
    TerrainType.BEACH: Color.CHAMPAGNE_MIST,
    TerrainType.GRASS: Color.LIGHT_GREEN,
    TerrainType.GRASSHILLS: Color.LIGHT_GREEN,
    TerrainType.PLAINS: Color.CAMEL,
    TerrainType.PLAINSHILLS: Color.CAMEL,
    TerrainType.MOUNTAINS: Color.ROSY_GRANITE,
    TerrainType.SNOWPEAK: Color.BRIGHT_SNOW,
    TerrainType.FOREST: Color.BLACK_FOREST,
    TerrainType.UNDEFINED: Color.HOT_PINK
}

class VisibilityLevel(CoopEnum):
    HIDDEN = auto()
    REVEALABLE = auto()
    REVEALING = auto()
    REVEALED = auto()

@dataclass(frozen=True)
class NeighborAttrs:
    shared_verts: Dict[int, vec.FloatVec]
    cnxn_node_id: str

    @property
    def Boundary(self)-> line.Line:
        if len(self.shared_verts) >= 2:
            return line.Line(*self.shared_verts.values())
        return None

@dataclass
class Region:
    name: UniqueIdentifier
    pt: vec.FloatVec2D
    terrain: TerrainType
    boundary: Polygon
    neighbors: Dict[int, NeighborAttrs]
    visibility_level: VisibilityLevel = field(init=False, default=VisibilityLevel.HIDDEN)
    action_states: List = field(default_factory=list, init=False)
    id: uuid.UUID = field(init=False, default_factory=uuid.uuid4)
    node_id: str = field(init=False)
    vert_node_ids: List[str]

    def __post_init__(self):
        self.node_id = f"R{self.name}"

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        if type(other) != Region:
            return False

        return self.__hash__() == other.__hash__()

    def add_action_state(self, state):
        self.action_states.append(state)
        self.action_states = list(set(self.action_states))
        return self

    def remove_action_state(self, state, force: bool = False):
        if state not in self.action_states and not force:
            raise Exception(f"{state} not in region \'{self.name}\'s action_states")

        if state in self.action_states:
            self.action_states.remove(state)

        return self

    def set_visibility(self, requested, force:bool = False):
        dependencies = {
            VisibilityLevel.REVEALABLE: [VisibilityLevel.HIDDEN, VisibilityLevel.REVEALABLE],
            VisibilityLevel.REVEALED: [VisibilityLevel.REVEALABLE, VisibilityLevel.REVEALING, VisibilityLevel.REVEALED],
            VisibilityLevel.HIDDEN: [VisibilityLevel.HIDDEN],
            VisibilityLevel.REVEALING: [VisibilityLevel.REVEALABLE, VisibilityLevel.REVEALING]
        }

        if self.visibility_level not in dependencies[requested] and not force:
            raise Exception(f"Cannot set {requested} a region that was {self.visibility_level}")

        self.visibility_level = requested

        return self

    @property
    def RiverEligible(self):
        if self.terrain not in [
            TerrainType.MOUNTAINS,
            TerrainType.SNOWPEAK,
            TerrainType.DEEPSEAWATER,
            TerrainType.SHALLOWSEAWATER
        ]:
            return True

        return False



from dataclasses import dataclass
from behaviorTypes import BehaviorType
from typing import Dict
import cooptools.typeProviders as tp

@dataclass
class InputController:
    villager_counts: Dict[BehaviorType, tp.IntProvider]



import time

import pygame
from typing import Callable, Tuple, Dict
from cooptools.toggles import BooleanToggleable, BooleanToggleableProvider
from typing import List, Any, Callable
from dataclasses import dataclass, field
from cooptools.register import Register
import logging
from cooptools import typeProviders as tp
from cooptools.decor import TimeableProtocol

logger = logging.getLogger(__name__)

@dataclass(frozen=True)
class SurfaceGroup:
    surfaces_lookup: Dict[str, pygame.Surface]
    surface_pos: Dict[str, Tuple[int, int]]

DrawCallback = Callable[[], pygame.Surface | SurfaceGroup]

@dataclass(frozen=True)
class SurfaceRegistryArgs:
    id: str
    callback: DrawCallback
    frame_update: bool = False
    default_visible: bool = True
    visibility_toggle_provider: BooleanToggleableProvider = None
    pos: Tuple[int, int] = field(default=(0, 0))

@dataclass
class SurfaceRegister:
    registry_args: SurfaceRegistryArgs
    visibility_toggle: BooleanToggleableProvider = field(init=False, default_factory=lambda: BooleanToggleable(default=True))
    _allow_toggle_vis: bool = True

    def __post_init__(self):
        if self.registry_args.default_visible == False:
            self.visibility_toggle.set_value(False)

        if self.registry_args.visibility_toggle_provider is not None:
            self.visibility_toggle = self.registry_args.visibility_toggle_provider
            self._allow_toggle_vis = False

class SurfaceManager:
    def __init__(self,
                 surface_draw_callbacks: List[SurfaceRegistryArgs] = None,
                 time_tracker: TimeableProtocol = None):
        self.surface_register: Register = Register[SurfaceRegister]()
        self._time_tracker = time_tracker

        if surface_draw_callbacks:
            self.register_surface_ids(surface_draw_callbacks)

        self.surfaces: Dict[str, pygame.Surface] = {}

    @property
    def RegisteredSurfaceIds(self) -> List[str]:
        return list(self.surface_register.Registry.keys())

    @property
    def FrameUpdateSurfaceIds(self) -> List[str]:
        return [surf.registry_args.id for id, surf in self.surface_register.Registry.items()
                if surf.registry_args.frame_update]

    def register_surface_ids(self,
                             surface_draw_callbacks: List[SurfaceRegistryArgs]):
        self.surface_register.register(
            to_register=[SurfaceRegister(x) for x in surface_draw_callbacks],
            ids=[x.id for x in surface_draw_callbacks]
        )
        logger.info(f"Registered surfaces: [{[x.id for x in surface_draw_callbacks]}]")
        return self

    def invalidate(self, ids: List[str] = None):
        if ids is not None:
            logger.info(f"Invalidating {ids}")
        else:
            logger.info(f"Invalidating")

        self.redraw(ids=ids)

    def redraw(self,
               ids: List[str] = None):
        if ids is None:
            ids = list(self.surface_register.Registry.keys())

        for id in ids:
            tic = time.perf_counter()
            logger.info(f"Drawing surface \'{id}\'...")
            self.surfaces[id] = self.surface_register.Registry[id].registry_args.callback()
            toc = time.perf_counter()
            logger.info(f"Surface \'{id}\' drawn [{round(toc - tic, 3) / 1000.0}s]")
            if self._time_tracker is not None:
                self._time_tracker.internally_tracked_times.setdefault(f"redraw_{id}", 0)
                self._time_tracker.internally_tracked_times[f"redraw_{id}"] += toc - tic

    def update(self):
        self.update_if_visible(self.FrameUpdateSurfaceIds)

    def render(self,
               surface: pygame.Surface,
               frame_update: bool = False,
               surface_group_ids_whitelist: Dict[str, List[str]] = None):
        if frame_update:
            self.update()

        for id, sr in self.surface_register.Registry.items():
            if tp.resolve(sr.visibility_toggle).value:
                to_render = self.get_surfaces(ids=[id])[id]

                # Handle a surface group where specific surfaceGroup ids are provided
                if (type(to_render) == SurfaceGroup and surface_group_ids_whitelist is not None
                        and id in surface_group_ids_whitelist.keys()):
                    logger.debug(f"Rendering surface group {id} [{surface_group_ids_whitelist[id]}]")
                    for to_render_sub_id in surface_group_ids_whitelist[id]:
                        surface.blit(to_render.surfaces_lookup[to_render_sub_id],
                                     dest=to_render.surface_pos[to_render_sub_id])

                # Handle a surface group where no specific surfaceGroup ids are provided
                if type(to_render) == SurfaceGroup and surface_group_ids_whitelist is None:
                    logger.debug(f"Rendering surface group {id}")
                    for id, surf in to_render.surfaces_lookup.items():
                        surface.blit(surf, dest=sr.registry_args.pos)

                # Handle if the type is not a surface group
                if type(to_render) == pygame.Surface:
                    logger.debug(f"Rendering surface {id}")
                    surface.blit(self.get_surfaces(ids=[id])[id], dest=sr.registry_args.pos)

    def get_surfaces(self,
                     ids: List[str],
                     force_update: bool = False) -> Dict[str, pygame.Surface | SurfaceGroup]:
        # update surfaces that dont exist or that dont match dims
        for id in ids:
            if id not in self.surfaces.keys():
                logger.info(f"Redrawing surface [{id}] because it was not cached")
                self.redraw([id])

            # elif type(self.surfaces[id]) == pygame.Surface and self.surfaces[id].get_size() != dims:
            #     logger.info(f"Redrawing surface [{id}] because it was not the correct size")
            #     self.redraw([id])

            elif force_update:
                logger.info(f"Redrawing surface [{id}] because it was forced")
                self.redraw([id])

        # return the surfaces
        return {id: self.surfaces[id] for id in ids}

    def get_toggled_state(self, ids: List[str]) -> Dict[str, bool]:
        return {
            x: tp.resolve(self.surface_register.Registry[x].visibility_toggle).value for x in ids
        }

    def toggle_visible(self, ids: List[str]) -> Dict[str, bool]:
        logger.info(f"Toggling visibility for {ids}")

        for x in ids:
            self.surface_register.Registry[x].visibility_toggle.toggle()

        return self.get_toggled_state(ids)

    def set_visiblility(self, ids: List[str], visible: bool):
        logger.info(f"Setting visibility [{visible}] for {ids}")

        for x in ids:
            self.surface_register.Registry[x].visibility_toggle.set_value(visible)

        return self.get_toggled_state(ids)

    def hide_all(self) -> Dict[str, bool]:
        logger.info(f"Hiding all surfaces")
        self.set_visiblility(ids=self.RegisteredSurfaceIds, visible=False)
        return self.get_toggled_state(self.RegisteredSurfaceIds)

    def show_all(self) -> Dict[str, bool]:
        logger.info(f"Showing all surfaces")
        self.set_visiblility(ids=self.RegisteredSurfaceIds, visible=True)
        return self.get_toggled_state(self.RegisteredSurfaceIds)

    def update_if_visible(self, ids: List[str]):
        self.redraw([id for id, visible in self.get_toggled_state(ids).items() if visible])




from coopgame.gameTemplate import GameTemplate, tt
from coopgame.surfaceManager import SurfaceManager, SurfaceRegistryArgs
import coopgame.pygamehelpers as help
import coopgame.handleKeyPressedArgs as hkpa
from voroni import *
from constants import *
from wave_collapse import *
from world import GameModel
from examples.voroni_world.draw_args import *
from cooptools import graphs as grf
from typing import Iterable, Tuple
import pygame_widgets
from coopgame.ui import *
from input_controller import InputController
from behaviorTypes import BehaviorType
import logging
from cooptools import common as cmn

logger = logging.getLogger(__name__)

class SurfaceLayer(CoopEnum):
    REGION_BORDERS = auto()
    REGION_COLORS = auto()
    REGION_LABELS = auto()
    MOUSE_REGION = auto()
    CNXNS = auto()
    POINTS = auto()
    VILLAGE = auto()
    VILLAGERS = auto()
    FOG_OF_WAR = auto()
    DEBUG_REGIONS = auto()
    USER_ROUTE = auto()
    RIVERS = auto()
    # PAUSED = auto()
    SECTOR_TREE = auto()
    INPUT_SURFACE_BACKGROUND = auto()
    INPUT_SURFACE = auto()
    GAME_INFO = auto()

class VoroniGame(GameTemplate):
    def __init__(self,
                 n_regions: int,
                 seed: int = None,
                 window_dims: Tuple[int, int] = None):
        super().__init__(input_state_handler=self.define_input_handler(),
                         screen_width=window_dims[0] if window_dims is not None else None,
                         screen_height=window_dims[1] if window_dims is not None else None
                         )
        self._surface_manager = SurfaceManager(time_tracker=tt)
        self.user_route_def = []
        self.user_route = None

        input_x_off = int(self.ui_side_size()[0] * .20)
        input_y_off = 50
        dims = (int(self.ui_side_size()[0] * .8), 10)
        inp_surf = help.init_surface(self.ui_side_size(), color_key=Color.HOT_PINK)
        self._scout_count_slider = TheSlider(
            surf=inp_surf,
            dims=dims,
            pos= (input_x_off, input_y_off),
            label="Scouts"
        )

        self._farmer_count_slider = TheSlider(
            surf=inp_surf,
            pos= (input_x_off, input_y_off + 60),
            label="Farmers",
            dims=dims,
        )
        self._scout_count_slider.surf.set_colorkey(Color.WHITE.value)

        ic = InputController(
            villager_counts={
                BehaviorType.SCOUT: lambda: self._scout_count_slider.Value,
                BehaviorType.FARM: lambda: self._farmer_count_slider.Value
            }
        )

        self._model = GameModel(input_controller=ic,
                                max_villagers=MAX_VILLAGERS,
                                init_villagers=INIT_VILLAGERS,
                                villager_spawn_ms=VILLAGER_SPAWN_MS,
                                sector_tree_depth=SECTOR_TREE_DEPTH,
                                seconds_per_hr=SECONDS_PER_HR)

        self._scout_count_slider.set_bounds(
            bounds=(0, lambda: self._model.NVillagers)
        )
        self._farmer_count_slider.set_bounds(
            bounds=(0, lambda: self._model.NVillagers)
        )

        self._surface_manager.register_surface_ids(
            surface_draw_callbacks=[
                SurfaceRegistryArgs(id=SurfaceLayer.REGION_COLORS.name, callback=self._get_surface_region_colors, pos=self.world_window_origin()),
                SurfaceRegistryArgs(id=SurfaceLayer.REGION_BORDERS.name, callback=self._get_surface_region_borders, pos=self.world_window_origin()),
                SurfaceRegistryArgs(id=SurfaceLayer.RIVERS.name, callback=self._get_river_surface, pos=self.world_window_origin()),
                SurfaceRegistryArgs(id=SurfaceLayer.VILLAGE.name, callback=self._get_village_surface, pos=self.world_window_origin()),
                SurfaceRegistryArgs(id=SurfaceLayer.POINTS.name, callback=self._get_surface_region_points, default_visible=False, pos=self.world_window_origin()),
                SurfaceRegistryArgs(id=SurfaceLayer.CNXNS.name, callback=self._get_surface_cnxns, default_visible=False, pos=self.world_window_origin()),
                SurfaceRegistryArgs(id=SurfaceLayer.MOUSE_REGION.name, callback=self._get_mouse_region_surface, pos=self.world_window_origin()),
                SurfaceRegistryArgs(id=SurfaceLayer.FOG_OF_WAR.name, callback=self._get_fow_surface),
                SurfaceRegistryArgs(id=SurfaceLayer.REGION_LABELS.name, callback=self._get_surface_region_names, default_visible=False, pos=self.world_window_origin()),
                SurfaceRegistryArgs(id=SurfaceLayer.VILLAGERS.name, callback=self._get_villager_surface, pos=self.world_window_origin()),
                SurfaceRegistryArgs(id=SurfaceLayer.DEBUG_REGIONS.name, callback=self._get_debug_surface, pos=self.world_window_origin()),
                SurfaceRegistryArgs(id=SurfaceLayer.USER_ROUTE.name, callback=self._get_user_route, pos=self.world_window_origin()),
                SurfaceRegistryArgs(id=SurfaceLayer.SECTOR_TREE.name, callback=self._get_sector_graph_surface, default_visible=False, pos=self.world_window_origin()),
                SurfaceRegistryArgs(id=SurfaceLayer.INPUT_SURFACE_BACKGROUND.name, callback=self._get_input_background_surface, default_visible=True),
                SurfaceRegistryArgs(id=SurfaceLayer.INPUT_SURFACE.name, callback=self._get_input_surface, default_visible=True),
                SurfaceRegistryArgs(id=SurfaceLayer.GAME_INFO.name, callback=self._get_game_info_surface, default_visible=True, pos=self.world_window_origin()),
            ]
        )



        self.n_regions = n_regions
        self.redefine(
            n=self.n_regions,
            seed=seed
        )

    def _handle_left_click(self):
        mouse_reg = self.get_mouse_region()

        if len(self.user_route_def) >= 2:
            self.user_route_def = []
        else:
            self.user_route_def.append(mouse_reg)

        logger.info([x.name for x in self.user_route_def])

        if len(self.user_route_def) == 2:
            self.user_route = self._model.route_between_regions(
                start=self.user_route_def[0],
                end=self.user_route_def[1]
            )
        else:
            self.user_route = None

        self._surface_manager.invalidate([SurfaceLayer.USER_ROUTE.name])

    def redefine(self, n: int, seed:int = None):

        self._model.redefine(
            # world_bounds=(self.screen_width, self.screen_height),
            world_bounds=self.world_window_size(),
            n_regions=n,
            seed=seed
        )
        self._surface_manager.invalidate()


    def ui_side_size(self):
        return (min(max(int(UI_SIDE_WIDTH_PERC * self.screen_width), UI_SIDE_WIDTH_MIN), UI_SIDE_WIDTH_MAX), self.screen_height)

    def world_window_size(self):
        return (self.screen_width - self.ui_side_size()[0], self.screen_height)

    def world_window_origin(self):
        return self.ui_side_size()[0], 0

    def define_input_handler(self):
        return (((
            hkpa.InputStateHandler(
                quit_callback_package=hkpa.CallbackPackage(down=lambda x: self.quit()),
                fullscreen_callback_package=hkpa.CallbackPackage(down=lambda x: self.toggle_fullscreen()),
                key_actions=[
                    hkpa.InputAction([hkpa.InputEvent(hkpa.PyKeys.K_r, event_type=hkpa.InputEventType.DOWN)],callback=lambda x: self.redefine()),
                    hkpa.InputAction([hkpa.InputEvent(hkpa.PyKeys.K_g, event_type=hkpa.InputEventType.DOWN)], callback=lambda x: self._surface_manager.toggle_visible([SurfaceLayer.CNXNS.name])),
                    hkpa.InputAction([hkpa.InputEvent(hkpa.PyKeys.K_KP_PLUS, event_type=hkpa.InputEventType.DOWN)], callback=lambda x: self._surface_manager.show_all()),
                    hkpa.InputAction([hkpa.InputEvent(hkpa.PyKeys.K_KP_MINUS, event_type=hkpa.InputEventType.DOWN)], callback=lambda x: self._surface_manager.hide_all()),
                    hkpa.InputAction([hkpa.InputEvent(hkpa.PyKeys.K_f, event_type=hkpa.InputEventType.DOWN)], callback=lambda x: self._surface_manager.toggle_visible([SurfaceLayer.FOG_OF_WAR.name])),
                    hkpa.InputAction([hkpa.InputEvent(hkpa.PyKeys.K_s, event_type=hkpa.InputEventType.DOWN)], callback=lambda x: self._surface_manager.toggle_visible([SurfaceLayer.SECTOR_TREE.name])),
                    hkpa.InputAction([hkpa.InputEvent(hkpa.PyMouse.LEFT, event_type=hkpa.InputEventType.DOWN)], callback=lambda x: self._handle_left_click()),
            ])
            .with_debug(lambda x: self.debug_mode_toggle.toggle()))
            .with_number_mapping(
                {
                    0: lambda x: self._surface_manager.toggle_visible([SurfaceLayer.REGION_BORDERS.name]),
                    1: lambda x: self._surface_manager.toggle_visible([SurfaceLayer.REGION_LABELS.name]),
                    2: lambda x: self._surface_manager.toggle_visible([SurfaceLayer.POINTS.name]),
                    3: lambda x: self._surface_manager.toggle_visible([SurfaceLayer.REGION_COLORS.name])
                }
            ))
            .with_pause(lambda x: self.paused_toggle.toggle()))

    def _draw_region(self,
                    surf,
                    r: Region,
                     border_color: Color = None,
                     background_color: Color = None,
                     pt_color: Color = None,
                     label_color: Color = None,
                     offset: Tuple[int, int] = None):
        if r is None:
            return

        if offset is None:
            offset = (0, 0)

        if background_color:
            # draw highlighted areas
            pygame.draw.polygon(surf,
                                background_color.value,
                                [(x + offset[0], y + offset[1]) for x, y in r.boundary.exterior.coords])

        if border_color is not None:
            ii = 0
            while ii < len(r.boundary.exterior.coords) - 1:
                start_pos = r.boundary.exterior.coords[ii]
                end_pos = r.boundary.exterior.coords[ii + 1]

                pygame.draw.line(surf,
                                 border_color.value,
                                 vec.add_vectors([offset, start_pos]),
                                 vec.add_vectors([offset, end_pos]),
                                 2)
                ii += 1

        if pt_color is not None:
            # draw points
            pnt.draw_points(
                {
                    (float(x.pt[0] + offset[0]), float(x.pt[1] + offset[1])): pnt.DrawPointArgs(
                        color=pt_color
                    ) for x in self._model.regions.values()
                },
                surface=surf
            )

        if label_color is not None:
            lbl.draw_label(hud=surf,
                           args=lbl.LabelArgs(
                               text=str(r.name),
                               draw_args=lbl.DrawLabelArgs(
                                   color=label_color,
                                   alignment=TextAlignmentType.CENTER
                               ),
                               pos=vec.add_vectors([r.pt, offset]),
                           ))

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def get_mouse_region(self):
        mouse_pos = help.mouse_pos()
        mouse_reg = self._model.closest_region(vec.vector_between(self.world_window_origin(), mouse_pos))

        return mouse_reg

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_surface_region_borders(self):
        if self._model.regions is None:
            return None

        surf = help.init_surface(self.screen.get_size())
        surf.fill(color=Color.BLACK.value)

        # draw all the edges
        for p in self._model.regions.values():
            self._draw_region(surf, p, border_color=VERONI_EDGECOLOR)

        return surf

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_surface_region_points(self):
        if self._model.regions is None:
            return None

        surf = help.init_surface(self.world_window_size())
        surf.fill(color=Color.BLACK.value)

        for p in self._model.regions.values():
            self._draw_region(
                surf=surf,
                pt_color=Color.RED,
                r=p
            )

        return surf

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_surface_region_colors(self):
        if self._model.regions is None:
            return None

        surf = help.init_surface(self.world_window_size())
        surf.fill(color=Color.BLACK.value)

        # draw all the edges
        for p in self._model.regions.values():
            self._draw_region(surf,
                              r=p,
                              background_color=TERRAIN_COLOR_MAP[p.terrain])
        return surf


    def _define_graph_drawer(self,
                             show_graph: bool = False,
                             routes: Iterable[grf.AStarResults] = None):
        return pgh.PyGraphHandler(
            screen=self.screen,
            draw_config=GRAPH_DRAW_CONFIG if show_graph else None,
            graph_getter=lambda: self._model.world_graph,
            route_draw_args_getter=lambda: [(r, ROUTE_DRAW_ARGS) for r in routes] if routes else None,
        )

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_surface_cnxns(self):
        if self._model.regions is None:
            return None

        surf = help.init_surface(self.world_window_size())
        surf.fill(color=Color.BLACK.value)

        self._define_graph_drawer(show_graph=True).render(surface=surf)
        return surf

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_surface_region_names(self):
        if self._model.regions is None:
            return None

        surf = help.init_surface(self.world_window_size())
        surf.fill(color=Color.BLACK.value)

        # draw labels
        for p in self._model.regions.values():
            self._draw_region(surf,
                              r=p,
                              label_color=Color.MAGENTA)

        return surf

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_village_surface(self):
        if self._model.regions is None:
            return None

        surf = help.init_surface(self.world_window_size())
        surf.fill(color=Color.BLACK.value)

        surf.set_alpha(120)

        vill = self._model.regions[self._model.village_id]

        # draw highlighted areas
        self._draw_region(
            surf=surf,
            r=vill,
            background_color=VILLAGE_COLOR
        )

        return surf

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_fow_surface(self):
        if self._model.regions is None:
            return None

        surf = pygame.image.load(r'C:\Users\Tj Burns\PycharmProjects\coopgame\examples\voroni_world\assets\imgs\fog_of_war.png')
        surf.set_colorkey(Color.BLACK.value)
        surf = pygame.transform.scale(surface=surf, size=self.screen.get_size())

        # draw a black (see through) area for the revealed region
        for p in self._model.regions.values():
            if p.visibility_level == VisibilityLevel.REVEALED:
                self._draw_region(
                    surf=surf,
                    r=p,
                    background_color=Color.BLACK,
                    offset=self.world_window_origin()
                )

            if p.visibility_level == VisibilityLevel.REVEALABLE:
                self._draw_region(
                    surf=surf,
                    r=p,
                    background_color=Color.PINK_ORCHID,
                    border_color=VERONI_EDGECOLOR,
                    offset=self.world_window_origin()
                    # label_color=Color.MAGENTA
                )
            if p.visibility_level == VisibilityLevel.REVEALING:
                self._draw_region(
                    surf=surf,
                    r=p,
                    background_color=Color.LIGHT_BLUE,
                    border_color=VERONI_EDGECOLOR,
                    offset=self.world_window_origin()
                    # label_color=Color.MAGENTA
                )
        return surf

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_game_info_surface(self):
        surf = help.init_surface(self.world_window_size())
        surf.fill(color=Color.BLACK.value)

        left_align = self.world_window_size()[0] - 200

        # Draw world clock
        days, hrs = self._model.world_time()
        lbl.draw_label(hud=surf,
                       args=lbl.LabelArgs(
                           text=f"Day {days}, {hrs}:00",
                           draw_args=lbl.DrawLabelArgs(
                               color=VILLAGER_COLOR,
                               alignment=TextAlignmentType.CENTER
                           ),
                           pos=(left_align, 0),
                       ))

        # Draw villager spawn clock
        lbl.draw_label(hud=surf,
                       args=lbl.LabelArgs(
                           text=f"Next Villager Spawn: {round(self._model.villager_spawn_timer.MsRemaining / 1000.0, 1)}s",
                           draw_args=lbl.DrawLabelArgs(
                               color=VILLAGER_COLOR,
                               alignment=TextAlignmentType.CENTER
                           ),
                           pos=(left_align, 40),
                       ))

        offset = 40
        delta = 25
        for name, villager in self._model.Villagers.items():
            offset += delta
            lbl.draw_label(hud=surf,
                           args=lbl.LabelArgs(
                               text=f"{villager.name} --> {tuple(f'{round(x, 1)}...' for x in villager.Position)}",
                               draw_args=lbl.DrawLabelArgs(
                                   color=VILLAGER_COLOR,
                                   alignment=TextAlignmentType.CENTER
                               ),
                               # pos=(self.screen_width * left_align_perc, self.screen_height * offset),
                               pos=(left_align,  offset)
                           ))

        return surf

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_villager_surface(self):
        if self._model.regions is None:
            return None

        surf = help.init_surface(self.world_window_size())
        surf.fill(color=Color.BLACK.value)

        # draw villager label and point
        for v in self._model.Villagers.values():
            lbl.draw_label(
                hud=surf,
                args=lbl.LabelArgs(
                    text=str(v.name),
                    draw_args=lbl.DrawLabelArgs(
                        color=VILLAGER_COLOR,
                    ),
                    pos=v.waypoint_follower.Transform.Translation.Vector,
                    anchor_point_draw_args=pnt.DrawPointArgs(
                        color=VILLAGER_COLOR,
                        radius=3
                    )
                )
            )

            # Draw arrow to villager final dest
            if ((v.waypoint_follower.NextDestination is not None)
                    and (v.waypoint_follower.Position != v.waypoint_follower.NextDestination.end_pos)):
                line.draw_lines(
                    {line.Line(
                        origin=v.waypoint_follower.Position,
                        destination=v.waypoint_follower.NextDestination.end_pos,
                    ): line.DrawLineArgs(
                        color=Color.SLATE_GRAY,
                        directionality_draw_args=line.DirectionalityIndicatorDrawArgs(
                            color=Color.SLATE_GRAY,
                            height=5,
                            width=5
                        ),

                    )},
                    surface=surf

                )

            # Draw arrow to villager next waypoint
            if ((v.waypoint_follower.CurrentTarget is not None)
                    and (v.waypoint_follower.Position != v.waypoint_follower.CurrentTarget.end_pos)):
                line.draw_lines(
                    {line.Line(
                        origin=v.waypoint_follower.Position,
                        destination=v.waypoint_follower.CurrentTarget.end_pos,
                    ): line.DrawLineArgs(
                        color=VILLAGER_COLOR,
                        directionality_draw_args=line.DirectionalityIndicatorDrawArgs(
                            color=VILLAGER_COLOR,
                            height=5,
                            width=5
                        ),

                    )},
                    surface=surf

                )

        return surf

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_mouse_region_surface(self):
        if self._model.regions is None:
            return None

        surf = help.init_surface(self.world_window_size())
        surf.fill(color=Color.BLACK.value)

        # draw labels
        region = self.get_mouse_region()

        self._draw_region(surf,
                          r=region,
                          label_color=Color.MAGENTA,
                          border_color=Color.YELLOW)

        return surf

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_user_route(self):
        if self.user_route is None:
            return None

        surf = help.init_surface(self.world_window_size())
        surf.fill(color=Color.BLACK.value)

        self._define_graph_drawer(routes=[self.user_route]).render(surface=surf)
        return surf

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_river_surface(self):
        if self._model.rivers is None:
            return None

        surf = help.init_surface(self.world_window_size())
        surf.fill(color=Color.BLACK.value)

        gh = pgh.PyGraphHandler(
            screen=self.screen,
            draw_config=RIVER_GRAPH_DRAW_CONFIG,
            graph_getter=lambda: self._model.rivers,
        )

        gh.render(surface=surf)

        return surf

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_debug_surface(self):
        if self._model.crash_regions is None:
            return None

        surf = help.init_surface(self.world_window_size())
        surf.fill(color=Color.BLACK.value)

        regions = self._model.crash_regions

        for region in regions:
            self._draw_region(surf,
                              r=region,
                              label_color=Color.MAGENTA,
                              border_color=Color.RED)
        help.draw_arrow(
            surf,
            start=regions[0].pt,
            end=regions[1].pt,
            color=Color.RED,
            leader_line=True
        )
        return surf

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_input_background_surface(self):
        surf = pygame.image.load(
            r'C:\Users\Tj Burns\PycharmProjects\coopgame\examples\voroni_world\assets\imgs\side_menu_background.png',
        ).convert_alpha()
        surf = pygame.transform.scale(surface=surf, size=self.ui_side_size())
        return surf

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_input_surface(self):
        self._scout_count_slider.update()
        self._farmer_count_slider.update()

        '''Update pygame widgets'''
        pygame_widgets.update(self._events)

        return self._scout_count_slider.surf


    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_sector_graph_surface(self):
        surf = help.init_surface(self.world_window_size())
        # surf.fill(color=Color.BLACK.value)

        sg = self._model.world_sector_map

        sectors_to_draw = [(pos, sector[0], sector[1]) for pos, sector in sg.Sectors.items()]
        sectors_to_draw.sort(key=lambda x: x[2], reverse=True)
        for pos, sect_rect, layer in sectors_to_draw:
            color_map = {
                0: (Color.DARK_BLUE, 3),
                1: (Color.DODGER_BLUE, 2),
                2: (Color.LIGHT_BLUE, 1)
            }

            help.draw_box(
                surface=surf,
                rect=help.Rectangle(pt=(sect_rect[0], sect_rect[1]),
                                    dims=(sect_rect[2], sect_rect[3]),
                                    ),
                width=color_map.get(layer, (0, 1))[1],
                color=color_map.get(layer, (Color.LIGHT_GRAY, ))[0]
            )

            if layer < 2:
                labutils.draw_label(
                    hud=surf,
                    args=labutils.LabelArgs(
                        text=str(pos),
                        pos=(sect_rect[0], sect_rect[1]),
                        draw_args=labutils.DrawLabelArgs(
                            color=color_map.get(layer, (Color.LIGHT_GRAY, ))[0]
                        )
                    ),
                )
        return surf



    def draw(self, frames: int, debug_mode: bool = False):
        self._surface_manager.invalidate([SurfaceLayer.VILLAGERS.name,
                                          SurfaceLayer.MOUSE_REGION.name,
                                          SurfaceLayer.DEBUG_REGIONS.name,
                                          SurfaceLayer.INPUT_SURFACE.name,
                                          SurfaceLayer.GAME_INFO.name])
        self._surface_manager.render(self.screen)
        pygame.display.set_caption(f'Voroni Game: seed {self._model._seed}')



    def on_resize(self):
        self.redefine(seed=self._model._seed)
        self._surface_manager.invalidate()


    def model_updater(self, delta_time_ms: int):
        results = self._model.update(delta_time_ms)

        revealed_regions = []
        if results is not None:
            revealed_regions = cmn.flattened_list_of_lists(x.revealed_regions for x in results)

        if revealed_regions is not None and len(revealed_regions) > 0:
            self._surface_manager.invalidate([SurfaceLayer.FOG_OF_WAR.name])

if __name__ == '__main__':
    import os
    from cooptools.loggingHelpers import BASE_LOG_FORMAT
    logging.basicConfig(format=BASE_LOG_FORMAT, level=logging.DEBUG)
    for log_provider in [
        'cooptools.sectors.sectorTree.sectorTree_SectorComparer',
        'cooptools.physics.kinematic',
        'coopgame.surfaceManager',
        'cooptools.timers',
        'cooptools.graphs.graph',
        'cooptools.graphs.graph_dcs',
        'cooptools.timeTracker.timeTracker',
        'cooptools.sectors.sectorTree.sectorTree'
    ]:
        logging.getLogger(log_provider).setLevel(logging.WARNING)

    for log_provider in [
        'cooptools.graphs.graph_dcs'
    ]:
        logging.getLogger(log_provider).setLevel(logging.INFO)

    def define_file_logger():
        # Ensure the logs directory exists
        if not os.path.exists('logs'):
            os.makedirs('logs')

        # 1. Get the specific logger instance
        logger = logging.getLogger('cooptools.sectors.sectorTree.sectorTree')

        # Set the logger's overall minimum level (handlers can filter further)
        logger.setLevel(logging.DEBUG)

        # 2. Create a FileHandler
        log_file_path = os.path.join('logs', 'app_events.log')
        if os.path.exists(log_file_path):
            os.remove(log_file_path)

        file_handler = logging.FileHandler(log_file_path, mode='a')  # 'a' for append, 'w' for write/overwrite

        # 3. Set the FileHandler's logging level
        # file_handler.setLevel(logging.INFO)  # This handler will only process INFO messages and above

        # 4. Create a Formatter
        formatter = logging.Formatter(
            BASE_LOG_FORMAT,
            datefmt='%Y-%m-%d %H:%M:%S'
        )

        # 5. Add Formatter to the Handler
        file_handler.setFormatter(formatter)

        # 6. Add the Handler to the Logger
        logger.addHandler(file_handler)

    # define_file_logger()
    logging.getLogger('__main__').setLevel(logging.NOTSET)
    g = VoroniGame(
        N_REGIONS,
        # seed=1694791782243067367
        # window_dims=(1500, 750),
    )
    g.run(debug=False)



from cooptools.colors import Color

UI_SIDE_WIDTH_PERC = .2
UI_SIDE_WIDTH_MIN = 200
UI_SIDE_WIDTH_MAX = 300
N_REGIONS = 100
VERONI_EDGECOLOR = Color.closest_color_to_rgb((100, 100, 100))
VERONI_PERC_TO_COLOR = .3
VERONI_FILLCOLOR = Color.BLUE
NEIGHBOR_CNXN_COLOR = Color.MAROON
VILLAGE_COLOR = Color.RED
VILLAGER_MAX_ACCEL = 5
VILLAGER_MAX_VELO = 10
VILLAGER_COLOR = Color.PURPLE
MAX_VILLAGERS = 2
INIT_VILLAGERS = 1
VILLAGER_SPAWN_MS = 30000
SECTOR_TREE_DEPTH = 4
SECONDS_PER_HR = 5


from typing import Protocol, Iterable, Self, Dict
from coopstructs.geometry import Rectangle
from cooptools.protocols import UniqueIdentifier
from cooptools.geometry_utils import vector_utils as vec
from cooptools.geometry_utils import polygon_utils as ply


class CollisionProtocol(Protocol):
    def check_collides(self, colliders: Dict[UniqueIdentifier, Self]) -> Dict[UniqueIdentifier, bool]:
        pass


class TwoDimensionConvexPolyCollider:
    def __init__(self,
                 poly: vec.IterVec):
        self.poly = poly

    def check_collides(self, colliders: Dict[UniqueIdentifier, Self]) -> Dict[UniqueIdentifier, bool]:
        ret = {}
        for id, collider in colliders:
            ret[id] = ply.do_convex_polygons_intersect(self.poly, collider.poly)
        return ret

if __name__ == "__main__":
    import math
    def test_01():
        c1 = TwoDimensionConvexPolyCollider(poly =Rectangle.from_meta(0, 0, 100, 200, math.pi/4).Corners)
        c2 = TwoDimensionConvexPolyCollider(poly =Rectangle.from_meta(0, 0, 100, 200, math.pi/4).Corners)


from coopgame.gameTemplate import GameTemplate
import coopgame.pygamehelpers as help
from cooptools.colors import Color
from coopgame.pointdrawing import point_draw_utils as pdu

class HelpTest(GameTemplate):

    def __init__(self):

        super().__init__()

    def draw(self, frames: int, debug_mode: bool = False):
        pdu.draw_points(
            surface=self.screen,
            points={
                self.ScreenCenter: pdu.DrawPointArgs(
                    color=Color.BLUE,
                    radius=50,
                    outline_color=Color.ORANGE
                )
            }
        )

        help.draw_arrow(self.screen,
                        self.ScreenCenter,
                        help.mouse_pos_as_vector().as_tuple(),
                        Color.GREEN,
                        leader_line=True,
                        arrow_height=20)


if __name__ == "__main__":
    g = HelpTest()
    g.main()


import uuid
from typing import Dict
from coopgame.gameTemplate import GameTemplate
from coopgame.handleKeyPressedArgs import InputStateHandler, InputAction, InputEvent, InputEventType, basic_key_actions, CallbackPackage
from coopgame.pygame_k_constant_names import PyKeys, PyMouse
from coopgame.linedrawing.pyCurveHandler import PyCurveHandler, CurveSurfaceType
from coopgame.linedrawing.curve_draw_utils import CurveDrawArgs
from coopgame.linedrawing.line_draw_utils import DrawLineArgs
from coopgame.pointdrawing.point_draw_utils import DrawPointArgs
from coopstructs.geometry.curves.curves import Curve, CubicBezier
from coopstructs.geometry.curves.curveBuilder import CurveBuilder
from coopstructs.geometry.rectangles.rectangle import Rectangle
import coopgame.pygamehelpers as help
from cooptools.toggles import BooleanToggleable
import cooptools.matrixManipulation as mm
import numpy as np
from cooptools.colors import Color
import cooptools.os_manip as osm
from coopui.cli.CliAtomicUserInteraction import CliAtomicUserInteraction as cli
from coopstructs.geometry.curves.curve_factory import curves_from_dict
from cooptools.common import flattened_list_of_lists
from coopstructs.geometry.curves.enums import CurveType

class CurveGame_Basic(GameTemplate):

    def __init__(self):
        super().__init__(input_state_handler=self.define_input_handler())

        self.curve_handler = PyCurveHandler(
            screen=self.screen,
            get_dynamic_curves_callback=self._get_dynamic_curves,
            get_static_curves_callback=self._get_static_curves
        )
        self._curve_builder = CurveBuilder(id_provider=lambda: str(uuid.uuid4()), default_curve_type=CurveType.CUBICBEZIER)
        self.use_scaler_matrix = BooleanToggleable(default=False)
        self.force_continuity = True
        self.show_arc_curve_approx = BooleanToggleable(default=True)


    def define_input_handler(self):
        return InputStateHandler(
            quit_callback_package=CallbackPackage(down=lambda x: self.quit()),
            debug_callback_package=CallbackPackage(down=lambda x: self.debug_mode_toggle.toggle()),
            fullscreen_callback_package=CallbackPackage(down=lambda x: self.toggle_fullscreen()),
            key_actions=[
                InputAction([InputEvent(PyKeys.K_r, event_type=InputEventType.DOWN)], callback=lambda x: self.curve_handler.redraw()),
                InputAction([InputEvent(PyKeys.K_0, event_type=InputEventType.DOWN)], callback=lambda x: self.curve_handler.toggle_surface([CurveSurfaceType.DYNAMIC_CURVES_ID])),
                InputAction([InputEvent(PyKeys.K_1, event_type=InputEventType.DOWN)], callback=lambda x: self.curve_handler.toggle_surface([CurveSurfaceType.STATIC_CURVES_ID])),
                InputAction([InputEvent(PyKeys.K_2, event_type=InputEventType.DOWN)], callback=lambda x: self.curve_handler.toggle_surface([CurveSurfaceType.DYNAMIC_OVERLAY_ID])),
                InputAction([InputEvent(PyKeys.K_3, event_type=InputEventType.DOWN)], callback=lambda x: self.curve_handler.toggle_surface([CurveSurfaceType.STATIC_OVERLAY_ID])),
                InputAction([InputEvent(PyMouse.LEFT, event_type=InputEventType.DOWN)],callback=lambda x: self.handle_left_mouse_down()),
                InputAction([InputEvent(PyMouse.RIGHT, event_type=InputEventType.DOWN)], callback=lambda x: self.handle_right_mouse_down()),
                InputAction([InputEvent(PyKeys.K_t, event_type=InputEventType.DOWN)], callback=lambda x: self.toggle_curve_type()),
                InputAction([InputEvent(PyKeys.K_s, event_type=InputEventType.DOWN)], callback=lambda x: self.save_curve()),
                InputAction([InputEvent(PyKeys.K_l, event_type=InputEventType.DOWN)], callback=lambda x: self.load_curves()),
                InputAction([InputEvent(PyKeys.K_c, event_type=InputEventType.DOWN)], callback=lambda x: self.show_arc_curve_approx.toggle()),
        ])

    def next_point(self):
        game_area_rect = Rectangle.from_tuple((0, 0, self.screen.get_height(), self.screen.get_width()))
        if self.force_continuity:
            return self._curve_builder.force_continuity(
                help.normal_mouse_position(game_area_rect=game_area_rect, draw_scale_matrix=self.draw_scale_matrix()))
        else:
            return help.normal_mouse_position(draw_scale_matrix=self.draw_scale_matrix(), game_area_rect=game_area_rect)

    def handle_left_mouse_down(self):
        self._curve_builder.add_point(self.next_point())
        self.curve_handler.invalidate()

    def handle_right_mouse_down(self):
        self._curve_builder.remove_point()
        self.curve_handler.invalidate()

    def _get_dynamic_curves(self) -> Dict[Curve,CurveDrawArgs]:
        new_curves = self._curve_builder.curves_from_temp_next_points([self.next_point()])

        curves = {curve: CurveDrawArgs(
            line_args=DrawLineArgs(
                color=Color.LIGHT_BLUE,
                width=2),
            control_line_args=DrawLineArgs(
                color=Color.GREY,
                width=1,
            ),
            control_point_args=DrawPointArgs(
                color=Color.RED,
                radius=5,
                outline_color=Color.YELLOW,
                outline_width=1
            ),
        )
            for curve in new_curves}

        if self.show_arc_curve_approx.value:
            arc_approx = flattened_list_of_lists([CubicBezier.as_circular_arcs(x, tolerance=5) for x in new_curves if type(x) == CubicBezier])

            curves = {**curves,
                **{x : CurveDrawArgs(
                        line_args=DrawLineArgs(
                            color=Color.RED,
                            width=2))
                    for x in arc_approx}}



        return curves

    def _get_static_curves(self) -> Dict[Curve,CurveDrawArgs]:
        curves = {curve: CurveDrawArgs(
            line_args=DrawLineArgs(
                color=Color.GREEN,
                width=2),
            buffer=5,
            buffer_color=Color.DARK_KHAKI
        )
            for curve in self._curve_builder.curves.values()}

        return curves


    def draw(self, frames: int, debug_mode: bool = False):
        self.curve_handler.render(self.screen)
        help.draw_points(
            {self.next_point().as_tuple(): DrawPointArgs(
                outline_color=Color.GREY,
                radius=5
            )},
            surface=self.screen
        )

        help.draw_mouse_pos(
            DrawPointArgs(
                color=Color.GREY,
                outline_color=Color.RED,
                radius=7
            ),
            screen=self.screen
        )



    def draw_scale_matrix(self):
        if self.use_scaler_matrix.value:
            return mm.rotateAroundPointMatrix((self.screen_width/2, self.screen_height / 2),  np.pi/4, np.pi/4)
        else:
            return None

    def toggle_curve_type(self):
        self._curve_builder.toggle_curve_type()

    def save_curve(self):
        self._curve_builder.save_curve()
        self.curve_handler.invalidate()

    def load_curves(self):
        fp = r'C:\Users\tburns\PycharmProjects\coopstructs\tests\testData\curves.json'

        if fp is None:
            fp=cli.request_open_filepath(prompt=f"Select a curves file", title="Curves", filetypes=((osm.FileType.JSON.value, ".json"), ))

        if fp is None:
            return

        data = osm.try_load_json_data_to_dict(file_path=fp)
        if data is None:
            return

        self._curve_builder.curves.clear()
        self._curve_builder.add_curves(curves_from_dict(data))
        self._curve_builder.save_curve()
        self.curve_handler.invalidate()


if __name__ == "__main__":
    g = CurveGame_Basic()
    g.main()


from coopgame.characterDevelopment import character as char
import coopgame.characterDevelopment.stats as stat
from typing import Iterable

def acquire_skills(character: char.Character, skills: Iterable[stat.StatEffectPackage]):
    character.stats.register_events(events=[stat.StatsPackageStateEvent(description="Add Skills", added_stat_modifiers=skills)])
    return character

def lose_skills(character: char.Character, skills: Iterable[stat.StatEffectPackage]):
    character.stats.register_events(events=[stat.StatsPackageStateEvent(description="Lose Skills", removed_stat_modifiers=skills)])
    return character



import coopgame.pygamehelpers as help
from typing import Iterable, Tuple
import pygame_widgets
from pygame_widgets.slider import Slider
from pygame_widgets.textbox import TextBox
from dataclasses import dataclass, field
import pygame
import logging
from cooptools.colors import Color
import cooptools.typeProviders as tp

logger = logging.getLogger(__name__)

@dataclass(slots=True)
class TheSlider:
    surf: pygame.Surface
    dims: Tuple[int, int]
    pos: Tuple[int, int]
    font_size: int = field(default=10)
    bounds: Tuple[int, int] = field(init=False, default=None)
    label: str = field(default=None)
    slider: Slider = field(init=False, default=None)
    txt: TextBox = field(init=False, default=None)
    lbl: TextBox = field(init=False, default=None)
    __cache_bounds: Tuple[int, int] = field(init=False, default=None)
    __label_slider_y_off: int = field(default=30)

    def __post_init__(self):
        self.lbl = TextBox(self.surf, self.pos[0] + 10, self.pos[1], 50, 20, fontSize=self.font_size)
        self.txt = TextBox(self.surf, self.pos[0] + 10, self.pos[1] + self.__label_slider_y_off, 20, 20, fontSize=self.font_size)

        self.txt.disable()
        self.lbl.disable()

    def __resolve_bounds(self):
        if self.bounds is None:
            raise Exception(f"Bounds have not been set. Call the set_bounds method after init")

        return (tp.resolve(self.bounds[0]), tp.resolve(self.bounds[1]))

    def __build_slider(self):
        txt_strt = self.txt.getX() + self.txt.getWidth()
        txt_strt_w_buffer = txt_strt + 10

        kwargs = {
            "win": self.surf,
            "x": txt_strt_w_buffer,
            "y": self.pos[1] + self.__label_slider_y_off,
            "width": self.dims[0] - txt_strt_w_buffer - 10,
            "height": self.dims[1],
            "min": self.__cache_bounds[0],
            "max": self.__cache_bounds[1],
            "step": 1,
        }

        if self.slider is not None:
            kwargs["initial"] = self.Value

        return Slider(**kwargs)

    def update(self):

        if ((cur_bnds:= self.__resolve_bounds()) != self.__cache_bounds) or self.__cache_bounds is None:
            self.__cache_bounds = cur_bnds
            self.slider = self.__build_slider()

        self.surf.fill((255, 255, 255))
        self.txt.setText(str(self.slider.getValue()))
        self.lbl.setText(str(self.label) if self.label is not None else "")

    def set_bounds(self, bounds: Tuple[tp.IntProvider, tp.IntProvider]):
        self.bounds = bounds

    @property
    def Value(self) -> int:
        return int(self.slider.getValue())


from coopgame.gameTemplate import GameTemplate
import coopgame.pygamehelpers as help
from cooptools.colors import Color
from coopgame.grids_graphs.pygridhandler import PyGridHandler
from cooptools.toggles import BooleanToggleable
from coopgame.handleKeyPressedArgs import InputState, InputStateHandler, InputAction, InputEvent, InputEventType, basic_key_actions, CallbackPackage
from coopgame.pygame_k_constant_names import PyKeys, PyMouse
from cooptools.sectors.grids import RectGrid
from cooptools.dictPolicies import TogglePolicy
from coopstructs.geometry.vectors.vectorN import Vector2
from coopgame.grids_graphs.pygridhandler import GridSurfaceType
from coopgame.surfaceManager import SurfaceManager, SurfaceRegistryArgs
from cooptools.coopEnum import CoopEnum, auto
import pygame

BACK_COLOR = Color.YELLOW


class SurfaceLayer(CoopEnum):
    BACK = auto()
    BG0 = auto()
    BG1 = auto()
    FG0 = auto()
    FG1 = auto()
    CLOUDS = auto()


class SurfaceManager_Basic(GameTemplate):

    def __init__(self):
        super().__init__(input_state_handler=self.define_input_handler())
        self._surface_manager = SurfaceManager()

        self._surface_manager.register_surface_ids(
            surface_draw_callbacks=[
                SurfaceRegistryArgs(id=SurfaceLayer.BACK.name, callback=self._get_surface_back),
                SurfaceRegistryArgs(id=SurfaceLayer.BG0.name, callback=self._get_surface_bg0)
            ]
        )


    def define_input_handler(self):
        return InputStateHandler(
            quit_callback_package=CallbackPackage(down=lambda x: self.quit()),
            debug_callback_package=CallbackPackage(down=lambda x: self.debug_mode_toggle.toggle()),
            fullscreen_callback_package=CallbackPackage(down=lambda x: self.toggle_fullscreen()),
            key_actions=[
                InputAction([InputEvent(PyKeys.K_r, event_type=InputEventType.DOWN)], callback=lambda x: self.redraw()),
                InputAction([InputEvent(PyKeys.K_0, event_type=InputEventType.DOWN)], callback=lambda x: self._surface_manager.toggle_visible([SurfaceLayer.BACK.name])),
                InputAction([InputEvent(PyKeys.K_1, event_type=InputEventType.DOWN)], callback=lambda x: self._surface_manager.toggle_visible([SurfaceLayer.BG0.name])),
                InputAction([InputEvent(PyKeys.K_2, event_type=InputEventType.DOWN)], callback=lambda x: self._surface_manager.toggle_visible([SurfaceLayer.BG1.name])),
                InputAction([InputEvent(PyKeys.K_KP_PLUS, event_type=InputEventType.DOWN)], callback=lambda x: self._surface_manager.show_all()),
                InputAction([InputEvent(PyKeys.K_KP_MINUS, event_type=InputEventType.DOWN)], callback=lambda x: self._surface_manager.hide_all()),
        ])


    def _get_surface_back(self):
        surf = help.init_surface(self.screen.get_size())
        surf.fill(color=BACK_COLOR.value)
        return surf

    def _get_surface_bg0(self):
        bg = pygame.image.load(r"C:\Users\tburns\PycharmProjects\coopgame\coopgame\assets\backgrounds\desert_spikes_dawn.png")
        return bg

    def draw(self, frames: int, debug_mode: bool = False):
        self._surface_manager.render(self.screen)

    def redraw(self):
        self._surface_manager.redraw()


if __name__ == "__main__":
    g = SurfaceManager_Basic()
    g.main()


from cooptools.colors import Color
import coopgame.grids_graphs.draw_graph_utils as utils
import coopgame.linedrawing.line_draw_utils as lutils
import coopgame.label_drawing.label_drawing_utils as labutils
import coopgame.pointdrawing.point_draw_utils as putils
from coopgame import grids_graphs as pgh

ARROW_SIZES = 3

GRAPH_DRAW_CONFIG = utils.GraphDrawArgs(
            node_draw_args=putils.DrawPointArgs(
                color=Color.DARK_BLUE,
                radius=2,
            ),
            enabled_edge_args=lutils.DrawLineArgs(
                color=Color.ORANGE,
                directionality_draw_args=lutils.DirectionalityIndicatorDrawArgs(
                    color=Color.ORANGE,
                    height=ARROW_SIZES,
                    width=ARROW_SIZES
                ),
                draw_label_args=labutils.DrawLabelArgs(
                    color=Color.WHEAT
                )
            ),
            disabled_edge_args=lutils.DrawLineArgs(
                color=Color.BROWN,
                directionality_draw_args=lutils.DirectionalityIndicatorDrawArgs(
                    color=Color.BROWN,
                    height=ARROW_SIZES,
                    width=ARROW_SIZES
                ),
                draw_label_args=labutils.DrawLabelArgs(
                    color=Color.BROWN
                )
            ),
            node_label_args=labutils.DrawLabelArgs(
                color=Color.WHEAT
            ),
            # articulation_points_args=putils.DrawPointArgs(
            #     outline_color=Color.PURPLE,
            #     radius=10,
            #     outline_width=3
            # ),
            sink_node_args=putils.DrawPointArgs(
                color=Color.RED,
                radius=10
            ),
            source_node_args=putils.DrawPointArgs(
                color=Color.GREEN,
                radius=10
            ),
            orphan_node_args=putils.DrawPointArgs(
                outline_color=Color.RED,
                radius=10
            )
        )

ROUTE_DRAW_ARGS = pgh.RouteDrawArgs(
                edge_draw_args=lutils.DrawLineArgs(
                    color=Color.TEAL,
                    width=2,
                    # draw_label_args=labutils.DrawLabelArgs(
                    #     color=Color.TEAL
                    # ),
                    directionality_draw_args=lutils.DirectionalityIndicatorDrawArgs(
                        color=Color.TEAL,
                        height=ARROW_SIZES,
                        width=ARROW_SIZES
                    )
                ),
                node_draw_args=putils.DrawPointArgs(
                    color=Color.TEAL,
                    radius=2
                ),
                # node_label_args=labutils.DrawLabelArgs(
                #     color=Color.TEAL
                # )
            )

RIVER_GRAPH_DRAW_CONFIG = utils.GraphDrawArgs(
            # node_draw_args=putils.DrawPointArgs(
            #     color=Color.DARK_BLUE,
            #     radius=2,
            # ),
            enabled_edge_args=lutils.DrawLineArgs(
                color=Color.DARK_BLUE,
                width=5
                # directionality_draw_args=lutils.DirectionalityIndicatorDrawArgs(
                #     color=Color.ORANGE,
                #     height=ARROW_SIZES,
                #     width=ARROW_SIZES
                # ),
                # draw_label_args=labutils.DrawLabelArgs(
                #     color=Color.WHEAT
                # )
            ),
            # disabled_edge_args=lutils.DrawLineArgs(
            #     color=Color.BROWN,
            #     directionality_draw_args=lutils.DirectionalityIndicatorDrawArgs(
            #         color=Color.BROWN,
            #         height=ARROW_SIZES,
            #         width=ARROW_SIZES
            #     ),
            #     draw_label_args=labutils.DrawLabelArgs(
            #         color=Color.BROWN
            #     )
            # ),
            # node_label_args=labutils.DrawLabelArgs(
            #     color=Color.WHEAT
            # ),
            # articulation_points_args=putils.DrawPointArgs(
            #     outline_color=Color.PURPLE,
            #     radius=10,
            #     outline_width=3
            # ),
            sink_node_args=putils.DrawPointArgs(
                color=Color.ORANGE,
                radius=10
            ),
            source_node_args=putils.DrawPointArgs(
                color=Color.GREEN,
                radius=10
            ),
            orphan_node_args=putils.DrawPointArgs(
                outline_color=Color.RED,
                radius=10
            )
        )



from pyquaternion import Quaternion
import numpy as np

class CoopQuaternion(Quaternion):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def rotate_all(self, points: np.array, rotation_point: np.array):
        return self._rotate_all_mat(points, rotation_point)

    def _rotate_all_mat(self, points: np.array, rotation_point: np.array):
        if len(rotation_point) < 4:
            rotation_point = np.append(rotation_point, 0)

        centre = rotation_point[:4]
        centered_points = points - centre

        rotated_points = np.transpose(self.transformation_matrix.dot(np.transpose(centered_points)))
        recentered_points = rotated_points + centre

        return recentered_points
        # return np.vstack(recentered_points, np.ones(recentered_points.shape[0]))


    def _rotate_all_iter(self, points: np.array, rotation_point: np.array):
        # https://math.stackexchange.com/questions/18382/quaternion-and-rotation-about-an-origin-and-an-arbitrary-axis-origin-help
        centre = rotation_point[:3]
        centered_points = points[:, :3] - centre

        rotated_list = []

        # TODO: find an optimization that doesnt require iterating on the list of points
        for ii in range(0, len(points)):
            point = self.rotate(centered_points[ii]) + centre
            rotated_list.append(tuple(point) + (1, ))

        return np.array(rotated_list)


import sys
import numpy
import random as rnd
import pygame
from scipy.spatial import Voronoi
from coopgame.gameTemplate import GameTemplate, tt
from cooptools.colors import Color
from cooptools.geometry_utils import vector_utils as vec
from coopgame.label_drawing import label_drawing_utils as lbl
from coopgame.pointdrawing import point_draw_utils as pnt
from coopgame.linedrawing import line_draw_utils as line
from coopgame.label_drawing.pyLabel import TextAlignmentType
import numpy as np
from collections import defaultdict
from shapely.geometry import Polygon
from cooptools.coopEnum import CoopEnum, auto
from dataclasses import dataclass, field
from cooptools.coopDataclass import BaseIdentifiedDataClass, UniqueIdentifier
from typing import List, Dict, Set, Tuple, Self, Iterator
from cooptools.decor import timer
import logging
logger = logging.getLogger(__name__)
@timer(logger=logger, log_level=0, time_tracking_class=tt)
def voronoi_polygons(voronoi, diameter):
    """Generate shapely.geometry.Polygon objects corresponding to the
    regions of a scipy.spatial.Voronoi object, in the order of the
    input points. The polygons for the infinite regions are large
    enough that all points within a distance 'diameter' of a Voronoi
    vertex are contained in one of the infinite polygons.

    """
    centroid = voronoi.points.mean(axis=0)

    # Mapping from (input point index, Voronoi point index) to list of
    # unit vectors in the directions of the infinite ridges starting
    # at the Voronoi point and neighbouring the input point.
    ridge_direction = defaultdict(list)
    for (p, q), rv in zip(voronoi.ridge_points, voronoi.ridge_vertices):
        u, v = sorted(rv)
        if u == -1:
            # Infinite ridge starting at ridge point with index v,
            # equidistant from input points with indexes p and q.
            t = voronoi.points[q] - voronoi.points[p] # tangent
            n = np.array([-t[1], t[0]]) / np.linalg.norm(t) # normal
            midpoint = voronoi.points[[p, q]].mean(axis=0)
            direction = np.sign(np.dot(midpoint - centroid, n)) * n
            ridge_direction[p, v].append(direction)
            ridge_direction[q, v].append(direction)

    for i, r in enumerate(voronoi.point_region):
        region = voronoi.regions[r]
        if -1 not in region:
            # Finite region.
            yield Polygon(voronoi.vertices[region])
            continue
        # Infinite region.
        inf = region.index(-1)              # Index of vertex at infinity.
        j = region[(inf - 1) % len(region)] # Index of previous vertex.
        k = region[(inf + 1) % len(region)] # Index of next vertex.
        if j == k:
            # Region has one Voronoi vertex with two ridges.
            dir_j, dir_k = ridge_direction[i, j]
        else:
            # Region has two Voronoi vertices, each with one ridge.
            dir_j, = ridge_direction[i, j]
            dir_k, = ridge_direction[i, k]

        # Length of ridges needed for the extra edge to lie at least
        # 'diameter' away from all Voronoi vertices.
        length = 2 * diameter / np.linalg.norm(dir_j + dir_k)

        # Polygon consists of finite part plus an extra edge.
        finite_part = voronoi.vertices[region[inf + 1:] + region[:inf]]
        extra_edge = [voronoi.vertices[j] + dir_j * length,
                      voronoi.vertices[k] + dir_k * length]
        yield Polygon(np.concatenate((finite_part, extra_edge)))

@timer(logger=logger, log_level=0, time_tracking_class=tt)
def build_voroni_neigbors_map(v: Voronoi):
    vert_to_point_map = {}

    # create a map of the point_region and which point it maps to
    region_point_map = {int(x): ii for ii, x in enumerate(v.point_region)}
    point_region_map = {ii: int(x) for ii, x in enumerate(v.point_region)}

    '''Error control if duplicate points were supplied'''
    seen = set()
    dupes = {int(x) for ii, x in enumerate(v.point_region) if x in seen or seen.add(x)}
    deb = {}
    for x in dupes:
        deb[x] = []
        [deb[x].append(ii) for ii, v in enumerate(v.point_region) if x == v]

    if len(dupes) > 0:
        raise ValueError(f"Input points resulted in shared regions: {deb}")

    """We want to know which points define areas that share verticies. Therefore, we first wish to create a
    map of the verticies and which points they are a member of for fast lookup.
    For each vert in each defined region, add the area that it maps to."""
    for ii, verticies in enumerate([x for x in v.regions]):
        if len(verticies) == 0:
            continue
        for vert in verticies:
            vert_to_point_map.setdefault(vert, set()).add(region_point_map[ii])

    neighbor_map = {}

    for vert, points in vert_to_point_map.items():
        if vert == -1: continue

        for point in points:
            neighbor_map.setdefault(point, set())
            for sub in points:
                if sub != point:
                    neighbor_map[point].add(sub)

    border_map = {}
    for point, neighbors in neighbor_map.items():
        for n in neighbors:
            my_vert_indicies = [x for x in v.regions[point_region_map[point]] if x != -1]
            n_vert_indicies = [x for x in v.regions[point_region_map[n]] if x != -1]

            shared = [x for x in my_vert_indicies if x in n_vert_indicies]

            if len(shared) > 2: raise Exception(f"Neighbors share more than 2 verts: {my_vert_indicies}, {n_vert_indicies}")

            shared_verts = [tuple(float(x) for x in v.vertices[x]) for x in shared]

            border_map[frozenset((point, n))] = {x: shared_verts[ii] for ii, x in enumerate(shared)}

    return neighbor_map, border_map


@timer(logger=logger, log_level=0, time_tracking_class=tt)
def generate_voronoi(points: vec.IterVec):
    """
    Randomly chooses various points within the x and y dimensions of the map.
    Then, uses SciPy to generate a voronoi diagram with them, and returns it.

    :return: SciPy voronoi diagram
    """
    point_arr = numpy.zeros([len(points), 2], numpy.float32)

    for i in range(len(points)):
        point_arr[i][0] = points[i][0]
        point_arr[i][1] = points[i][1]

    v = Voronoi(point_arr)

    # print(point_arr[161], point_arr[749])
    vn_map, border_map = build_voroni_neigbors_map(v)
    polys = [x for x in voronoi_polygons(v, diameter=1000)]

    return v, vn_map, point_arr, polys, border_map



from cooptools.coopEnum import CoopEnum, add_succ_and_pred_maps, auto

@add_succ_and_pred_maps
class Orientation(CoopEnum):
    UP=1
    RIGHT=2
    DOWN=3
    LEFT=4

@add_succ_and_pred_maps
class ConnectionType(CoopEnum):
    ARC = 1
    LINE = 2
    BEZIER = 3

@add_succ_and_pred_maps
class EntryType(CoopEnum):
    CRAWLER = 1
    GRIDCLICK = 2

@add_succ_and_pred_maps
class GridDrawType(CoopEnum):
    BOXES = 1
    LINES = 2

@add_succ_and_pred_maps
class Rarity(CoopEnum):
    COMMON = auto()
    UNCOMMON = auto()
    RARE = auto()
    EPIC = auto()
    LEGENDARY = auto()
    MYTHIC = auto()
    DIVINE = auto()
